from dataclasses import dataclass
from typing import Callable

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin

__all__ = ["BERTConfig", "BERT"]


@dataclass
class BERTConfig:
    vocab_size: int
    hidden_size: int
    num_attention_heads: int
    num_hidden_layers: int
    intermediate_size: int

    hidden_act: Callable[[Tensor], Tensor] | str
    hidden_dropout_prob: float
    attention_probs_dropout_prob: float

    max_position_embeddings: int
    tie_word_embedding: bool
    type_vocab_size: int

    initializer_range: float
    layer_norm_eps: float

    use_cache: bool
    is_decoder: bool
    add_cross_attention: bool
    chunk_size_feed_forward: int

    pad_token_id: int = 0
    bos_token_id: int | None = None
    eos_token_id: int | None = None

    classifier_dropout: float | None = None
    add_pooling_layer: bool = True

    @classmethod
    def base(
        cls,
        *,
        is_decoder: bool = False,
        use_cache: bool = False,
        add_pooling_layer: bool = True,
        vocab_size: int = 30522,
        **kwargs,
    ) -> BERTConfig:
        defaults = dict(
            vocab_size=vocab_size,
            hidden_size=768,
            num_attention_heads=12,
            num_hidden_layers=12,
            intermediate_size=3072,
            hidden_act=F.gelu,
            hidden_dropout_prob=0.1,
            attention_probs_dropout_prob=0.1,
            max_position_embeddings=512,
            tie_word_embedding=True,
            type_vocab_size=2,
            initializer_range=0.02,
            layer_norm_eps=1e-12,
            use_cache=use_cache,
            is_decoder=is_decoder,
            add_cross_attention=False,
            chunk_size_feed_forward=0,
            pad_token_id=0,
            classifier_dropout=None,
            add_pooling_layer=add_pooling_layer,
        )
        defaults.update(kwargs)
        return cls(**defaults)

    @classmethod
    def large(
        cls,
        *,
        is_decoder: bool = False,
        use_cache: bool = False,
        add_pooling_layer: bool = True,
        vocab_size: int = 30522,
        **kwargs,
    ) -> BERTConfig:
        defaults = dict(
            vocab_size=vocab_size,
            hidden_size=1024,
            num_attention_heads=16,
            num_hidden_layers=24,
            intermediate_size=4096,
            hidden_act=F.gelu,
            hidden_dropout_prob=0.1,
            attention_probs_dropout_prob=0.1,
            max_position_embeddings=512,
            tie_word_embedding=True,
            type_vocab_size=2,
            initializer_range=0.02,
            layer_norm_eps=1e-12,
            use_cache=use_cache,
            is_decoder=is_decoder,
            add_cross_attention=False,
            chunk_size_feed_forward=0,
            pad_token_id=0,
            classifier_dropout=None,
            add_pooling_layer=add_pooling_layer,
        )
        defaults.update(kwargs)
        return cls(**defaults)


class _BERTEmbeddings(nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.word_embeddings = nn.Embedding(
            config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id
        )
        self.position_embeddings = nn.Embedding(
            config.max_position_embeddings, config.hidden_size
        )
        self.token_type_embeddings = nn.Embedding(
            config.type_vocab_size, config.hidden_size
        )

        self.layernorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

        self.position_ids: nn.Buffer
        self.token_type_ids: nn.Buffer
        self.register_buffer(
            "position_ids",
            nn.Buffer(
                lucid.arange(config.max_position_embeddings, dtype=lucid.Long).expand(
                    1, -1
                )
            ),
        )
        self.register_buffer(
            "token_type_ids",
            nn.Buffer(lucid.zeros(*self.position_ids.shape, dtype=lucid.Long)),
        )

    def forward(
        self,
        input_ids: lucid.LongTensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
        past_key_values_length: int = 0,
        cache_position: Tensor | None = None,
    ) -> Tensor:
        if input_ids is not None:
            input_shape = input_ids.shape
        else:
            input_shape = inputs_embeds.shape[:-1]

        batch_size, seq_length = input_shape
        if position_ids is None:
            if cache_position is not None:
                if cache_position.ndim != 1 or cache_position.shape[0] != seq_length:
                    raise ValueError(
                        "cache_position must be 1-D with length == seq_length "
                        f"(got shape={cache_position.shape}, seq_length={seq_length})."
                    )
                position_ids = cache_position.reshape(1, seq_length)
            else:
                position_ids = self.position_ids[
                    :, past_key_values_length : seq_length + past_key_values_length
                ]
        elif (
            position_ids.ndim != 2
            or position_ids.shape[1] != seq_length
            or position_ids.shape[0] not in (1, batch_size)
        ):
            raise ValueError(
                "position_ids must be 2-D with shape "
                "[batch_or_1, seq_length] where batch_or_1 is 1 or batch_size "
                f"(got shape={position_ids.shape}, batch_size={batch_size}, "
                f"seq_length={seq_length})."
            )

        if token_type_ids is None:
            if hasattr(self, "token_type_ids"):
                buffered_token_type_ids = self.token_type_ids.expand(
                    position_ids.shape[0], -1
                )
                buffered_token_type_ids = lucid.gather(
                    buffered_token_type_ids, dim=1, index=position_ids
                )
                token_type_ids = buffered_token_type_ids.expand(batch_size, seq_length)
            else:
                token_type_ids = lucid.zeros(
                    *input_shape, dtype=lucid.Long, device=self.position_ids.device
                )

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)

        token_type_embeddings = self.token_type_embeddings(token_type_ids)
        embeddings = inputs_embeds + token_type_embeddings

        position_embeddings = self.position_embeddings(position_ids)
        embeddings = embeddings + position_embeddings

        embeddings = self.layernorm(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


class _BERTSelfAttention(nn.Module):
    def __init__(
        self,
        config: BERTConfig,
        /,
        layer_idx: int | None = None,
    ) -> None:
        super().__init__()
        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                f"hidden_size {config.hidden_size} is not a multiple "
                f"of num_attention_heads {config.num_attention_heads}."
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)

        self.all_head_size = self.num_attention_heads * self.attention_head_size
        self.scaling = self.attention_head_size**-0.5

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

        self.is_decoder = config.is_decoder
        self.layer_idx = layer_idx

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        cache_position: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        input_shape = hidden_states.shape[:-1]
        hidden_shape = (*input_shape, -1, self.attention_head_size)

        query_layer = self.query(hidden_states).reshape(*hidden_shape).swapaxes(1, 2)
        key_layer = self.key(hidden_states).reshape(*hidden_shape).swapaxes(1, 2)
        value_layer = self.value(hidden_states).reshape(*hidden_shape).swapaxes(1, 2)

        if past_key_values is not None:
            if self.layer_idx is None:
                raise ValueError(f"past_key_values requires layer_idx")

            current_past_key_value = past_key_values
            if isinstance(past_key_values, nn.EncoderDecoderCache):
                current_past_key_value = past_key_values.self_attention_cache

            key_layer, value_layer = current_past_key_value.update(
                key_layer,
                value_layer,
                layer_idx=self.layer_idx,
                cache_position=cache_position,
            )

        attn_output = F.scaled_dot_product_attention(
            query_layer,
            key_layer,
            value_layer,
            attn_mask=attention_mask,
            dropout_p=self.dropout.p if self.training else 0.0,
            is_causal=False,
            scale=self.scaling,
        )
        attn_output = attn_output.swapaxes(1, 2).reshape(
            *input_shape, self.all_head_size
        )
        return attn_output


class _BERTCrossAttention(nn.Module):
    def __init__(
        self,
        config: BERTConfig,
        /,
        is_causal: bool = False,
        layer_idx: int | None = None,
    ) -> None:
        super().__init__()
        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                f"hidden_size '{config.hidden_size}' must be a "
                f"multiple of num_attention_heads '{config.num_attention_heads}'."
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = config.num_attention_heads * self.attention_head_size
        self.scaling = self.attention_head_size**-0.5

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        self.key = nn.Linear(config.hidden_size, self.all_head_size)
        self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)

        self.is_causal = is_causal
        self.layer_idx = layer_idx

    def forward(
        self,
        hidden_states: Tensor,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.EncoderDecoderCache | None = None,
        **kwargs,
    ) -> Tensor:
        bsz, tgt_len = hidden_states.shape[:-1]
        src_len = (
            encoder_hidden_states.shape[1]
            if encoder_hidden_states is not None
            else None
        )
        q_input_shape = (bsz, tgt_len, -1, self.attention_head_size)

        if past_key_values is not None and self.layer_idx is None:
            raise ValueError(f"past_key_values requires layer_idx")

        query_layer = self.query(hidden_states).reshape(*q_input_shape).swapaxes(1, 2)

        is_updated = (
            past_key_values.is_updated.get(self.layer_idx)
            if past_key_values is not None
            else False
        )
        if past_key_values is not None and is_updated:
            cached = past_key_values.cross_attention_cache.get(self.layer_idx)
            if cached is None:
                raise ValueError(
                    f"cross-attention cache is missing for layer_idx={self.layer_idx}."
                )
            key_layer, value_layer = cached
            src_len = key_layer.shape[-2]

        else:
            if encoder_hidden_states is None:
                raise ValueError(
                    "encoder_hidden_states is required when cache is empty."
                )
            src_len = encoder_hidden_states.shape[1]
            kv_input_shape = (bsz, src_len, -1, self.attention_head_size)

            key_layer = (
                self.key(encoder_hidden_states).reshape(*kv_input_shape).swapaxes(1, 2)
            )
            value_layer = (
                self.value(encoder_hidden_states)
                .reshape(*kv_input_shape)
                .swapaxes(1, 2)
            )
            if past_key_values is not None:
                key_layer, value_layer = past_key_values.update(
                    key_layer,
                    value_layer,
                    layer_idx=self.layer_idx,
                    is_cross_attention=True,
                )

        attn_output = F.scaled_dot_product_attention(
            query_layer,
            key_layer,
            value_layer,
            attn_mask=attention_mask,
            dropout_p=self.dropout.p if self.training else 0.0,
            scale=self.scaling,
            is_causal=self.is_causal,
        )
        attn_output = attn_output.swapaxes(1, 2).reshape(
            bsz, tgt_len, self.all_head_size
        )
        return attn_output


class _BERTSelfOutput(nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.layernorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states: Tensor, input_: Tensor) -> Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.layernorm(hidden_states + input_)

        return hidden_states


class _BERTAttention(nn.Module):
    def __init__(
        self,
        config: BERTConfig,
        /,
        is_causal: bool = False,
        layer_idx: int | None = None,
        is_cross_attention: bool = False,
    ) -> None:
        super().__init__()
        self.is_cross_attention = is_cross_attention
        if is_cross_attention:
            self.self = _BERTCrossAttention(
                config, is_causal=is_causal, layer_idx=layer_idx
            )
        else:
            self.self = _BERTSelfAttention(config, layer_idx=layer_idx)
        self.output = _BERTSelfOutput(config)

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        cache_position: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        attention_mask = (
            attention_mask if not self.is_cross_attention else encoder_attention_mask
        )
        if (
            self.is_cross_attention
            and past_key_values is not None
            and not isinstance(past_key_values, nn.EncoderDecoderCache)
        ):
            raise TypeError(
                "Cross-attention requires past_key_values to be "
                "nn.EncoderDecoderCache."
            )

        attn_output = self.self(
            hidden_states,
            encoder_hidden_states=encoder_hidden_states,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            cache_position=cache_position,
            **kwargs,
        )
        attn_output = self.output(attn_output, hidden_states)
        return attn_output


class _BERTIntermediate(nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.intermediate_size)
        self.intermediate_act_fn = (
            config.hidden_act
            if callable(config.hidden_act)
            else nn.utils.get_activation_from_name(config.hidden_act)
        )
        if self.intermediate_act_fn is None:
            raise ValueError(f"Invalid config.hidden_act name '{config.hidden_act}'")

    def forward(self, hidden_states: Tensor) -> Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class _BERTOutput(nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.layernorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states: Tensor, input_: Tensor) -> Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.layernorm(hidden_states + input_)

        return hidden_states


class _BERTLayer(nn.Module):
    def __init__(self, config: BERTConfig, layer_idx: int | None = None) -> None:
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1

        self.attention = _BERTAttention(
            config, is_causal=config.is_decoder, layer_idx=layer_idx
        )
        self.is_decoder = config.is_decoder
        self.add_cross_attention = config.add_cross_attention

        if self.add_cross_attention:
            if not self.is_decoder:
                raise ValueError(
                    f"{self} should be used as a decoder if cross attention is added."
                )
            self.crossattention = _BERTAttention(
                config, is_causal=False, layer_idx=layer_idx, is_cross_attention=True
            )

        self.intermediate = _BERTIntermediate(config)
        self.output = _BERTOutput(config)

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        cache_position: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        self_attention_output = self.attention(
            hidden_states,
            attention_mask,
            past_key_values=past_key_values,
            cache_position=cache_position,
            **kwargs,
        )
        attention_output = self_attention_output

        if self.is_decoder and encoder_hidden_states is not None:
            if not hasattr(self, "crossattention"):
                raise RuntimeError(
                    "If encoder_hidden_states is passed, "
                    "config.add_cross_attention must be set to True."
                )

            cross_attention_output = self.crossattention(
                self_attention_output,
                attention_mask=None,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                past_key_values=past_key_values,
                **kwargs,
            )
            attention_output = cross_attention_output

        layer_output = nn.utils.apply_chunking_to_forward(
            self.forward_chunk,
            self.chunk_size_feed_forward,
            self.seq_len_dim,
            attention_output,
        )
        return layer_output

    def forward_chunk(self, attention_output: Tensor) -> Tensor:
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)

        return layer_output


class _BERTEncoder(nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList(
            [_BERTLayer(config, layer_idx=i) for i in range(config.num_hidden_layers)]
        )

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: lucid.FloatTensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        use_cache: bool | None = None,
        cache_position: Tensor | None = None,
        **kwargs,
    ) -> Tensor:
        effective_use_cache = (
            use_cache if use_cache is not None else self.config.use_cache
        )
        if not self.config.is_decoder:
            effective_use_cache = False

        if (
            self.config.add_cross_attention
            and past_key_values is not None
            and not isinstance(past_key_values, nn.EncoderDecoderCache)
        ):
            raise TypeError(
                "When add_cross_attention=True, past_key_values must be "
                "nn.EncoderDecoderCache."
            )

        if not effective_use_cache:
            past_key_values = None
            cache_position = None
        elif past_key_values is None:
            raise ValueError(
                "use_cache=True requires a persistent past_key_values cache "
                "instance from the caller."
            )

        for layer_module in self.layer:
            hidden_states = layer_module(
                hidden_states,
                attention_mask,
                encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                past_key_values=past_key_values,
                cache_position=cache_position,
                **kwargs,
            )

        return hidden_states


class _BERTPooler(nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states: Tensor) -> Tensor:
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)

        return pooled_output


class BERT(PreTrainedModelMixin, nn.Module):
    def __init__(self, config: BERTConfig) -> None:
        super().__init__()
        self.config = config

        self.embeddings = _BERTEmbeddings(config)
        self.encoder = _BERTEncoder(config)
        self.pooler = _BERTPooler(config) if config.add_pooling_layer else None
        self.output_embeddings: nn.Linear | None = None

        self.apply(self._init_weights)
        self.tie_weights()

    def _init_weights(self, module: nn.Module) -> None:
        if isinstance(module, nn.Linear):
            nn.init.normal(module.weight, mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                nn.init.constant(module.bias, 0.0)

        elif isinstance(module, nn.Embedding):
            nn.init.normal(module.weight, mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx] = 0

        elif isinstance(module, nn.LayerNorm):
            nn.init.constant(module.bias, 0.0)
            nn.init.constant(module.weight, 1.0)

    def get_input_embeddings(self) -> nn.Embedding:
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value: nn.Embedding) -> None:
        self.embeddings.word_embeddings = value
        self.tie_weights()

    def get_output_embeddings(self) -> nn.Linear | None:
        return self.output_embeddings

    def set_output_embeddings(self, value: nn.Linear) -> None:
        self.output_embeddings = value
        self.tie_weights()

    def tie_weights(self) -> None:
        if not self.config.tie_word_embedding:
            return

        output_embeddings = self.get_output_embeddings()
        if output_embeddings is None:
            return

        output_embeddings.weight = self.get_input_embeddings().weight

    def _to_4d_attention_mask(
        self,
        attention_mask: Tensor | None,
        target_length: int,
        source_length: int,
        device: str,
    ) -> Tensor | None:
        if attention_mask is None:
            return None

        if attention_mask.device != device:
            attention_mask = attention_mask.to(device)

        if attention_mask.ndim == 2:
            if attention_mask.shape[1] != source_length:
                raise ValueError(
                    f"attention_mask has invalid source length: "
                    f"{attention_mask.shape[1]} != {source_length}"
                )
            attention_mask = attention_mask[:, None, None, :]
        elif attention_mask.ndim == 3:
            if attention_mask.shape[-1] != source_length:
                raise ValueError(
                    f"attention_mask has invalid source length: "
                    f"{attention_mask.shape[-1]} != {source_length}"
                )
            attention_mask = attention_mask[:, None, :, :]
        elif attention_mask.ndim == 4:
            if attention_mask.shape[-1] != source_length:
                raise ValueError(
                    f"attention_mask has invalid source length: "
                    f"{attention_mask.shape[-1]} != {source_length}"
                )
        else:
            raise ValueError(
                f"attention_mask must be 2-D, 3-D, or 4-D (got {attention_mask.ndim})."
            )

        if attention_mask.shape[-2] not in (1, target_length):
            raise ValueError(
                f"attention_mask has invalid target length axis: "
                f"{attention_mask.shape[-2]} not in (1, {target_length})"
            )

        attention_mask = attention_mask.astype(lucid.Float32)
        mask_min = float(lucid.min(attention_mask).item())
        mask_max = float(lucid.max(attention_mask).item())

        if mask_min >= 0.0 and mask_max <= 1.0:
            return (1.0 - attention_mask) * -1e12
        if mask_max <= 0.0:
            return attention_mask

        raise ValueError(
            "attention_mask values must be binary (0/1) or additive (<= 0 values)."
        )

    def _build_decoder_causal_mask(
        self,
        *,
        batch_size: int,
        target_length: int,
        source_length: int,
        device: str,
        past_key_values_length: int,
        position_ids: Tensor | None = None,
        cache_position: Tensor | None = None,
    ) -> Tensor:
        if position_ids is not None:
            query_positions = position_ids

        elif cache_position is not None:
            if cache_position.ndim != 1 or cache_position.shape[0] != target_length:
                raise ValueError(
                    "cache_position must be 1-D with length == target_length "
                    f"(got shape={cache_position.shape}, target_length={target_length})."
                )
            query_positions = cache_position.reshape(1, target_length)

        else:
            query_positions = lucid.arange(
                past_key_values_length,
                past_key_values_length + target_length,
                device=device,
            )
            query_positions = query_positions.reshape(1, target_length)

        if query_positions.ndim != 2 or query_positions.shape[1] != target_length:
            raise ValueError(
                "query positions must be 2-D with shape [batch_or_1, target_length] "
                f"(got shape={query_positions.shape}, target_length={target_length})."
            )

        if query_positions.shape[0] == 1:
            query_positions = query_positions.expand(batch_size, target_length)
        elif query_positions.shape[0] != batch_size:
            raise ValueError(
                "query positions batch axis must be 1 or batch_size "
                f"(got {query_positions.shape[0]} vs {batch_size})."
            )

        if query_positions.device != device:
            query_positions = query_positions.to(device)

        query_positions = query_positions.reshape(batch_size, target_length, 1)
        key_positions = lucid.arange(source_length, device=device).reshape(1, 1, -1)

        causal_mask = (key_positions > query_positions).astype(lucid.Float32) * -1e12
        return causal_mask[:, None, :, :]

    def forward(
        self,
        input_ids: lucid.LongTensor | None = None,
        attention_mask: Tensor | None = None,
        token_type_ids: lucid.LongTensor | None = None,
        position_ids: lucid.LongTensor | None = None,
        inputs_embeds: lucid.FloatTensor | None = None,
        encoder_hidden_states: lucid.FloatTensor | None = None,
        encoder_attention_mask: Tensor | None = None,
        past_key_values: nn.KVCache | nn.EncoderDecoderCache | None = None,
        use_cache: bool | None = None,
        cache_position: Tensor | None = None,
    ) -> tuple[Tensor, Tensor | None]:
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("Only one of input_ids or inputs_embeds can be provided.")

        if input_ids is None and inputs_embeds is None:
            raise ValueError("Either input_ids or inputs_embeds must be provided.")

        if input_ids is not None:
            input_shape = input_ids.shape
            device = input_ids.device
        else:
            input_shape = inputs_embeds.shape[:-1]
            device = inputs_embeds.device

        if len(input_shape) != 2:
            raise ValueError(
                f"BERT expects 2-D token inputs [batch, seq] (got {input_shape})."
            )

        effective_use_cache = (
            use_cache if use_cache is not None else self.config.use_cache
        )
        if not self.config.is_decoder:
            effective_use_cache = False

        if not effective_use_cache:
            past_key_values = None
            cache_position = None

        past_key_values_length = 0
        if effective_use_cache and past_key_values is not None:
            if isinstance(past_key_values, nn.EncoderDecoderCache):
                self_cache = past_key_values.self_attention_cache
                if len(self_cache.key_cache) > 0:
                    past_key_values_length = self_cache.get_seq_length()
            else:
                if len(past_key_values.key_cache) > 0:
                    past_key_values_length = past_key_values.get_seq_length()

        embedding_output = self.embeddings(
            input_ids=input_ids,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
            cache_position=cache_position,
        )

        source_length = input_shape[1] + past_key_values_length
        if attention_mask is None:
            attention_mask = lucid.ones((input_shape[0], source_length), device=device)

        elif attention_mask.ndim == 2 and attention_mask.shape[1] == input_shape[1]:
            if past_key_values_length > 0:
                past_attention = lucid.ones(
                    (input_shape[0], past_key_values_length), device=device
                )
                attention_mask = lucid.concatenate(
                    [past_attention, attention_mask], axis=1
                )

        extended_attention_mask = self._to_4d_attention_mask(
            attention_mask=attention_mask,
            target_length=input_shape[1],
            source_length=source_length,
            device=device,
        )
        if self.config.is_decoder:
            causal_attention_mask = self._build_decoder_causal_mask(
                batch_size=input_shape[0],
                target_length=input_shape[1],
                source_length=source_length,
                device=device,
                past_key_values_length=past_key_values_length,
                position_ids=position_ids,
                cache_position=cache_position,
            )
            if extended_attention_mask is None:
                extended_attention_mask = causal_attention_mask
            else:
                extended_attention_mask = (
                    extended_attention_mask + causal_attention_mask
                )

        if encoder_hidden_states is None:
            if encoder_attention_mask is not None:
                raise ValueError(
                    "encoder_attention_mask requires encoder_hidden_states."
                )
            extended_encoder_attention_mask = None
        else:
            extended_encoder_attention_mask = self._to_4d_attention_mask(
                attention_mask=encoder_attention_mask,
                target_length=input_shape[1],
                source_length=encoder_hidden_states.shape[1],
                device=device,
            )

        sequence_output = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=extended_encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            cache_position=cache_position,
        )
        pooled_output = (
            self.pooler(sequence_output) if self.pooler is not None else None
        )

        return sequence_output, pooled_output



import functools
import weakref
from typing import Any, Callable, overload

import lucid
from lucid._jit.tracer import TracingContext, _trace_local
from lucid._jit.passes import (
    run_passes,
    DEFAULT_INFERENCE_PASSES,
    DEFAULT_TRAINING_PASSES,
)
from lucid._jit.executor import (
    CompiledPlan,
    ForwardExecutor,
    ForwardResult,
    BackwardExecutor,
)
from lucid._jit.cache import PlanCache, CacheKey
from lucid._tensor.tensor import Tensor
from lucid._backend.core import BackwardOperation
from lucid._jit.ir import IRGraph


def _collect_tensor_inputs(args: tuple, kwargs: dict) -> tuple:
    result = []
    for a in args:
        if isinstance(a, Tensor):
            result.append(a)
    for v in kwargs.values():
        if isinstance(v, Tensor):
            result.append(v)
    return tuple(result)


def _trace_and_compile(
    fn: Callable,
    args: tuple,
    kwargs: dict,
    key: CacheKey,
    param_tensors: tuple = (),
) -> CompiledPlan:
    ctx = TracingContext()
    inputs = _collect_tensor_inputs(args, kwargs)

    for t in inputs:
        ctx.register_tensor(t, is_input=True)
    for p in param_tensors:
        ctx.register_tensor(p, is_param=True)

    _trace_local.active = True
    _trace_local.tracer = ctx
    try:
        outputs = fn(*args, **kwargs)
    finally:
        _trace_local.active = False
        _trace_local.tracer = None

    outputs_tuple = outputs if isinstance(outputs, tuple) else (outputs,)
    graph = ctx.finalize(outputs_tuple)

    training = key.grad_enabled and key.training_mode
    pass_classes = DEFAULT_TRAINING_PASSES if training else DEFAULT_INFERENCE_PASSES
    graph = run_passes(graph, [p() for p in pass_classes])

    return CompiledPlan(graph, training=training)


def _attach_compiled_backward(
    outputs: tuple,
    fwd_result: ForwardResult,
    graph: IRGraph,
) -> None:
    leaf_tensors = [
        fwd_result.value_map[vid]
        for vid in fwd_result.leaf_vids
        if fwd_result.value_map.get(vid) is not None
        and getattr(fwd_result.value_map[vid], "requires_grad", False)
    ]
    leaf_refs = tuple(weakref.ref(t) for t in leaf_tensors)
    leaf_versions = tuple(t._version for t in leaf_tensors)

    for out_tensor, out_vid in zip(outputs, graph.output_ids):
        if not getattr(out_tensor, "requires_grad", False):
            continue

        out_ref = weakref.ref(out_tensor)

        def _make_closure(_out_ref, _out_vid, _fwd):
            def _closure():
                _out = _out_ref()
                if _out is None or _out.grad is None:
                    return
                retain = getattr(_out, "keep_grad", False)
                BackwardExecutor().execute(
                    _fwd,
                    output_vids=[_out_vid],
                    output_upstream_grads={_out_vid: _out.grad},
                    retain_grad=retain,
                )

            return _closure

        out_tensor._prev = leaf_tensors
        out_tensor._backward_op = BackwardOperation(
            forward_op_ref=None,
            grad_func=None,
            tensor_refs=leaf_refs,
            versions=leaf_versions,
            device=None,
            custom_closure=_make_closure(out_ref, out_vid, fwd_result),
        )


class JITFunction:
    def __init__(
        self,
        fn: Callable,
        *,
        max_cache_entries: int = 8,
    ) -> None:
        self._fn = fn
        self._cache = PlanCache(max_entries=max_cache_entries)
        self._executor = ForwardExecutor()
        functools.update_wrapper(self, fn)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        inputs = _collect_tensor_inputs(args, kwargs)
        key = CacheKey.from_inputs(
            inputs,
            grad_enabled=lucid.grad_enabled(),
            training_mode=False,
        )
        plan = self._cache.get(key)

        if plan is None:
            plan = _trace_and_compile(self._fn, args, kwargs, key)
            self._cache.put(key, plan)

        return self._run_plan(plan, inputs, {}, lucid.grad_enabled())

    def _run_plan(
        self,
        plan: CompiledPlan,
        inputs: tuple,
        param_map: dict,
        grad_enabled: bool,
    ) -> Any:
        outputs, fwd_result = self._executor.execute(
            plan, inputs, param_map, grad_enabled
        )
        if fwd_result is not None:
            _attach_compiled_backward(outputs, fwd_result, plan.graph)
        return outputs[0] if len(outputs) == 1 else outputs

    def invalidate_cache(self) -> None:
        self._cache.invalidate()

    def __repr__(self) -> str:
        return f"JITFunction({self._fn.__name__}, cache_size={len(self._cache)})"


class JITModule:
    def __init__(
        self,
        module: Any,
        *,
        max_cache_entries: int = 8,
    ) -> None:
        self._module = module
        self._cache = PlanCache(max_entries=max_cache_entries)
        self._executor = ForwardExecutor()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        module = self._module
        inputs = _collect_tensor_inputs(args, kwargs)
        key = CacheKey.from_inputs(
            inputs,
            grad_enabled=lucid.grad_enabled(),
            training_mode=module.training,
        )

        plan = self._cache.get(key)
        if plan is None:
            param_tensors = tuple(module.parameters())
            plan = _trace_and_compile(
                module.forward, args, kwargs, key, param_tensors=param_tensors
            )
            self._cache.put(key, plan)

        for hook, with_kwargs in module._forward_pre_hooks:
            if with_kwargs:
                result = hook(module, args, kwargs)
                if result is not None:
                    args, kwargs = result
            else:
                result = hook(module, args)
                if result is not None:
                    args = result

        param_map = self._build_param_map(plan.graph)
        outputs, fwd_result = self._executor.execute(
            plan, inputs, param_map, lucid.grad_enabled()
        )
        output = outputs[0] if len(outputs) == 1 else outputs

        if fwd_result is not None:
            _attach_compiled_backward(outputs, fwd_result, plan.graph)

        for hook, with_kwargs in module._forward_hooks:
            if with_kwargs:
                result = hook(module, args, kwargs, output)
            else:
                result = hook(module, args, output)
            if result is not None:
                output = result

        return output

    def _build_param_map(self, graph) -> dict:
        result = {}
        for vid in graph.param_ids:
            iv = graph.values.get(vid)
            if iv is not None and iv.live_tensor is not None:
                result[vid] = iv.live_tensor
        return result

    def invalidate_cache(self) -> None:
        self._cache.invalidate()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._module, name)

    def __repr__(self) -> str:
        return (
            f"JITModule({type(self._module).__name__}, "
            f"cache_size={len(self._cache)})"
        )


@overload
def compile(
    target: Callable,
    *,
    max_cache_entries: int = 8,
) -> JITFunction: ...


@overload
def compile(
    target: Any,
    *,
    max_cache_entries: int = 8,
) -> JITModule: ...


def compile(
    target: Any,
    *,
    max_cache_entries: int = 8,
) -> JITFunction | JITModule:
    from lucid.nn.module import Module

    if isinstance(target, Module):
        return JITModule(target, max_cache_entries=max_cache_entries)
    elif callable(target):
        return JITFunction(target, max_cache_entries=max_cache_entries)
    else:
        raise TypeError(
            f"lucid.compile() expects an nn.Module or a callable, "
            f"got {type(target)}"
        )



import re

from collections import Counter
from functools import lru_cache
from pathlib import Path
from typing import Any, Iterable

import lucid

from lucid.data.tokenizers import Tokenizer, SpecialTokens
from lucid.data.tokenizers.utils import basic_tokenize, clean_text
from lucid.types import _DeviceType

from lucid._backend._C.tokenizers.core import _C_BPETokenizer, _C_ByteBPETokenizer

__all__ = [
    "BPETokenizer",
    "BPETokenizerFast",
    "ByteBPETokenizer",
    "ByteBPETokenizerFast",
]


class BPETokenizer(Tokenizer):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        merges: list[tuple[str, str]] | None = None,
        vocab_file: Path | str | None = None,
        merges_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        bos_token: SpecialTokens | str | None = None,
        eos_token: SpecialTokens | str | None = None,
        lowercase: bool = True,
        clean_text: bool = True,
        end_of_word_suffix: str = "</w>",
    ) -> None:
        if vocab is not None and vocab_file is not None:
            raise ValueError("Provide only one of 'vocab' or 'vocab_file'.")
        if merges is not None and merges_file is not None:
            raise ValueError("Provide only one of 'merges' or 'merges_file'.")

        super().__init__(unk_token, pad_token, bos_token, eos_token)
        if vocab_file is not None:
            vocab = self._load_vocab(vocab_file)
        if merges_file is not None:
            merges = self._load_merges(merges_file)

        self.lowercase = lowercase
        self.clean_text = clean_text
        self.end_of_word_suffix = end_of_word_suffix

        self.vocab: dict[str, int] = dict(vocab or {})
        self.merges: list[tuple[str, str]] = list(merges or [])
        self.merge_ranks: dict[tuple[str, str], int] = {
            pair: i for i, pair in enumerate(self.merges)
        }

        self._ensure_special_tokens()
        self.ids_to_tokens = self._build_id_to_token_map(self.vocab, verbose=False)

    @property
    def vocab_size(self) -> int:
        return len(self.vocab)

    def fit(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        verbose: bool = False,
    ) -> BPETokenizer:
        if min_frequency < 1:
            raise ValueError(f"'min_frequency' must be >= 1.")
        special_tokens = self.all_special_tokens
        if vocab_size < len(special_tokens):
            raise ValueError(
                "'vocab_size' must be >= "
                f"number of special tokens ({len(special_tokens)})."
            )

        word_freq = self._build_word_frequency(texts)
        if not word_freq:
            raise ValueError("Cannot train BPE vocabulary from empty corpus.")

        word_splits: dict[str, list[str]] = {}
        token_set: set[str] = set()

        for word in word_freq:
            chars = list(word)
            if not chars:
                continue
            chars[-1] = chars[-1] + self.end_of_word_suffix
            word_splits[word] = chars
            token_set.update(chars)

        merges: list[tuple[str, str]] = []
        target_size = vocab_size - len(special_tokens)

        while len(token_set) < target_size:
            pair_counts: Counter[tuple[str, str]] = Counter()
            for word, split in word_splits.items():
                freq = word_freq[word]
                for i in range(len(split) - 1):
                    pair_counts[(split[i], split[i + 1])] += freq

            if not pair_counts:
                break

            best_pair, best_count = pair_counts.most_common(1)[0]
            if best_count < min_frequency:
                break

            merges.append(best_pair)
            merged_token = best_pair[0] + best_pair[1]
            token_set.add(merged_token)

            for word, split in word_splits.items():
                i = 0
                new_split: list[str] = []
                while i < len(split):
                    if (
                        i < len(split) - 1
                        and split[i] == best_pair[0]
                        and split[i + 1] == best_pair[1]
                    ):
                        new_split.append(merged_token)
                        i += 2
                    else:
                        new_split.append(split[i])
                        i += 1

                word_splits[word] = new_split

        vocab_tokens = sorted(token_set)[:target_size]
        full_vocab = special_tokens + vocab_tokens

        self.vocab = {tok: i for i, tok in enumerate(full_vocab)}
        self.merges = merges
        self.merge_ranks = {pair: i for i, pair in enumerate(self.merges)}
        self.ids_to_tokens = self._build_id_to_token_map(self.vocab, verbose=verbose)

        return self

    @classmethod
    def train_from_iterator(
        cls,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        verbose: bool = False,
        **kwargs: Any,
    ) -> BPETokenizer:
        tok = cls(**kwargs)
        tok.fit(
            texts=texts,
            vocab_size=vocab_size,
            min_frequency=min_frequency,
            verbose=verbose,
        )
        return tok

    def tokenize(self, text: str) -> list[str]:
        if self.clean_text:
            text = clean_text(text)
        basic_tokens = basic_tokenize(text, lowercase=self.lowercase)

        out: list[str] = []
        for tok in basic_tokens:
            out.extend(self._bpe_tokenize(tok))
        return out

    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]:
        unk_id = self._unk_id()
        if isinstance(tokens, str):
            return self.vocab.get(tokens, unk_id)
        return [self.vocab.get(t, unk_id) for t in tokens]

    def convert_ids_to_tokens(self, ids: int | list[int]) -> str | list[str]:
        if isinstance(ids, int):
            return self._id_to_token(ids)
        return [self._id_to_token(i) for i in ids]

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        text = "".join(tokens)
        text = text.replace(self.end_of_word_suffix, " ")
        for punct in [".", ",", "!", "?", ";", ":", "%", ")", "]", "}"]:
            text = text.replace(f" {punct}", punct)
        for punct in ["(", "[", "{"]:
            text = text.replace(f"{punct} ", punct)

        text = re.sub(r"\s+", " ", text).strip()
        text = re.sub(r"([.!?])\1+", r"\1", text)

        return text

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        save_path = Path(save_directory)
        save_path.mkdir(parents=True, exist_ok=True)

        vocab_path = save_path / "vocab.json"
        merges_path = save_path / "merges.txt"

        self._save_json(self.vocab, vocab_path)
        with merges_path.open("w", encoding="utf-8") as f:
            for a, b in self.merges:
                f.write(f"{a} {b}\n")

        config = {
            "unk_token": self.unk_token,
            "pad_token": self.pad_token,
            "bos_token": self.bos_token,
            "eos_token": self.eos_token,
            "lowercase": self.lowercase,
            "clean_text": self.clean_text,
            "end_of_word_suffix": self.end_of_word_suffix,
        }
        config_path = self._save_tokenizer_config(save_path, config)
        return [str(vocab_path), str(merges_path), str(config_path)]

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: str | Path, **kwargs: Any
    ) -> BPETokenizer:
        path = Path(pretrained_model_name_or_path)
        if path.is_dir():
            base_dir = path
            vocab_file = base_dir / "vocab.json"
            merges_file = base_dir / "merges.txt"
        else:
            base_dir = path.parent
            vocab_file = path
            if vocab_file.name != "vocab.json":
                raise ValueError(
                    "Expected a directory containing 'vocab.json'/'merges.txt' "
                    "or a direct path to 'vocab.json'."
                )
            merges_file = base_dir / "merges.txt"

        if not vocab_file.exists():
            raise FileNotFoundError(f"Cannot find vocabulary file: {vocab_file}")
        if not merges_file.exists():
            raise FileNotFoundError(f"Cannot find merges file: {merges_file}")

        allowed = {
            "unk_token",
            "pad_token",
            "bos_token",
            "eos_token",
            "lowercase",
            "clean_text",
            "end_of_word_suffix",
        }
        init_kwargs = cls._load_tokenizer_config(base_dir, allowed_keys=allowed)
        init_kwargs.update(kwargs)

        return cls(vocab_file=vocab_file, merges_file=merges_file, **init_kwargs)

    @staticmethod
    def _load_vocab(vocab_file: Path | str) -> dict[str, int]:
        data = Tokenizer._load_json(vocab_file)
        out: dict[str, int] = {}
        for k, v in data.items():
            out[str(k)] = int(v)
        return out

    @staticmethod
    def _load_merges(merges_file: Path | str) -> list[tuple[str, str]]:
        merges: list[tuple[str, str]] = []
        with Path(merges_file).open("r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                parts = line.split()

                if len(parts) != 2:
                    continue
                merges.append((parts[0], parts[1]))

        return merges

    def _ensure_special_tokens(self) -> None:
        for tok in self.all_special_tokens:
            if tok not in self.vocab:
                self.vocab[tok] = len(self.vocab)

    def _unk_id(self) -> int:
        if self.unk_token is None or self.unk_token not in self.vocab:
            raise ValueError("Unknown token is not in vocabulary.")
        return self.vocab[self.unk_token]

    def _id_to_token(self, idx: int) -> str:
        if idx < 0 or idx >= len(self.ids_to_tokens):
            return self.unk_token or SpecialTokens.UNK.value

        t = self.ids_to_tokens[idx]
        return t if t else (self.unk_token or SpecialTokens.UNK.value)

    def _bpe_tokenize(self, token: str) -> list[str]:
        if not token:
            return []

        symbols = list(token)
        symbols[-1] = symbols[-1] + self.end_of_word_suffix

        while len(symbols) > 1:
            pairs = [(symbols[i], symbols[i + 1]) for i in range(len(symbols) - 1)]
            ranked = [(self.merge_ranks.get(p, 10**18), p) for p in pairs]

            best_rank, best_pair = min(ranked, key=lambda x: x[0])
            if best_rank == 10**18:
                break

            merged = best_pair[0] + best_pair[1]
            new_symbols: list[str] = []
            i = 0
            while i < len(symbols):
                if (
                    i < len(symbols) - 1
                    and symbols[i] == best_pair[0]
                    and symbols[i + 1] == best_pair[1]
                ):
                    new_symbols.append(merged)
                    i += 2
                else:
                    new_symbols.append(symbols[i])
                    i += 1

            symbols = new_symbols
        return symbols

    def _build_word_frequency(self, texts: Iterable[str]) -> Counter[str]:
        freq: Counter[str] = Counter()
        for text in texts:
            if self.clean_text:
                text = clean_text(text)
            for tok in basic_tokenize(text, lowercase=self.lowercase):
                if tok:
                    freq[tok] += 1
        return freq


class BPETokenizerFast(Tokenizer):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        merges: list[tuple[str, str]] | None = None,
        vocab_file: Path | str | None = None,
        merges_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        bos_token: SpecialTokens | str | None = None,
        eos_token: SpecialTokens | str | None = None,
        lowercase: bool = True,
        clean_text: bool = True,
        end_of_word_suffix: str = "</w>",
    ) -> None:
        if vocab is not None and vocab_file is not None:
            raise ValueError("Provide only one of 'vocab' or 'vocab_file'.")
        if merges is not None and merges_file is not None:
            raise ValueError("Provide only one of 'merges' or 'merges_file'.")

        super().__init__(unk_token, pad_token, bos_token, eos_token)

        backend_vocab_file: Path | None = None
        backend_merges_file: Path | None = None
        if vocab_file is not None:
            vocab = BPETokenizer._load_vocab(vocab_file)
        else:
            backend_vocab_file = None
        if merges_file is not None:
            merges = BPETokenizer._load_merges(merges_file)
        else:
            backend_merges_file = None

        self.lowercase = lowercase
        self.clean_text = clean_text
        self.end_of_word_suffix = end_of_word_suffix

        self.vocab: dict[str, int] = dict(vocab or {})
        self.merges: list[tuple[str, str]] = list(merges or [])

        self._backend = _C_BPETokenizer(
            vocab=vocab,
            merges=merges,
            vocab_file=backend_vocab_file,
            merges_file=backend_merges_file,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
            bos_token=self.bos_token,
            eos_token=self.eos_token,
            lowercase=self.lowercase,
            clean_text=self.clean_text,
            end_of_word_suffix=self.end_of_word_suffix,
        )
        self._sync_state_from_backend()

    @property
    def vocab_size(self) -> int:
        size = getattr(self._backend, "vocab_size")
        return int(size() if callable(size) else size)

    def tokenize(self, text: str) -> list[str]:
        return list(self._backend.tokenize(text))

    def encode(
        self,
        text: str,
        add_special_tokens: bool = True,
        return_tensor: bool = False,
        device: _DeviceType = "cpu",
    ) -> list[int] | lucid.LongTensor:
        ids = [int(x) for x in self._backend.encode_ids(text, add_special_tokens)]
        if return_tensor:
            return lucid.LongTensor(ids, device=device)
        return ids

    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]:
        if isinstance(tokens, str):
            return int(self._backend.convert_token_to_id(tokens))
        return [int(x) for x in self._backend.convert_tokens_to_ids(tokens)]

    def convert_ids_to_tokens(self, ids: int | list[int]) -> str | list[str]:
        if isinstance(ids, int):
            return str(self._backend.convert_id_to_token(ids))
        return [str(x) for x in self._backend.convert_ids_to_tokens(ids)]

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        return str(self._backend.convert_tokens_to_string(tokens))

    def fit(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
    ) -> BPETokenizerFast:
        self._backend.fit(list(texts), int(vocab_size), int(min_frequency))
        self._sync_state_from_backend()
        return self

    @classmethod
    def train_from_iterator(
        cls,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        **kwargs: Any,
    ) -> BPETokenizerFast:
        tokenizer = cls(**kwargs)
        tokenizer.fit(
            texts=texts,
            vocab_size=vocab_size,
            min_frequency=min_frequency,
        )
        return tokenizer

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        save_path = Path(save_directory)
        save_path.mkdir(parents=True, exist_ok=True)

        vocab_path = save_path / "vocab.json"
        merges_path = save_path / "merges.txt"

        if not self.vocab:
            self._sync_state_from_backend()

        self._save_json(self.vocab, vocab_path)
        with merges_path.open("w", encoding="utf-8") as f:
            for a, b in self.merges:
                f.write(f"{a} {b}\n")

        config = {
            "unk_token": self.unk_token,
            "pad_token": self.pad_token,
            "bos_token": self.bos_token,
            "eos_token": self.eos_token,
            "lowercase": self.lowercase,
            "clean_text": self.clean_text,
            "end_of_word_suffix": self.end_of_word_suffix,
        }
        config_path = self._save_tokenizer_config(save_path, config)
        return [str(vocab_path), str(merges_path), str(config_path)]

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str | Path,
        **kwargs: Any,
    ) -> BPETokenizerFast:
        path = Path(pretrained_model_name_or_path)
        if path.is_dir():
            base_dir = path
            vocab_file = base_dir / "vocab.json"
            merges_file = base_dir / "merges.txt"
        else:
            base_dir = path.parent
            vocab_file = path
            if vocab_file.name != "vocab.json":
                raise ValueError(
                    "Expected a directory containing 'vocab.json'/'merges.txt' "
                    "or a direct path to 'vocab.json'."
                )
            merges_file = base_dir / "merges.txt"

        if not vocab_file.exists():
            raise FileNotFoundError(f"Cannot find vocabulary file: {vocab_file}")
        if not merges_file.exists():
            raise FileNotFoundError(f"Cannot find merges file: {merges_file}")

        allowed = {
            "unk_token",
            "pad_token",
            "bos_token",
            "eos_token",
            "lowercase",
            "clean_text",
            "end_of_word_suffix",
        }
        init_kwargs = cls._load_tokenizer_config(base_dir, allowed_keys=allowed)
        init_kwargs.update(kwargs)

        return cls(vocab_file=vocab_file, merges_file=merges_file, **init_kwargs)

    def _sync_state_from_backend(self) -> None:
        backend_vocab = self._backend.get_vocab()
        self.vocab = {str(k): int(v) for k, v in backend_vocab.items()}

        backend_merges = self._backend.get_merges()
        self.merges = [(str(a), str(b)) for a, b in backend_merges]


class ByteBPETokenizer(BPETokenizer):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        merges: list[tuple[str, str]] | None = None,
        vocab_file: Path | str | None = None,
        merges_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        bos_token: SpecialTokens | str | None = None,
        eos_token: SpecialTokens | str | None = None,
        lowercase: bool = False,
        clean_text: bool = True,
        add_prefix_space: bool = False,
        end_of_word_suffix: str = "",
    ) -> None:
        loaded_from_artifacts = (
            vocab is not None
            or merges is not None
            or vocab_file is not None
            or merges_file is not None
        )
        self.add_prefix_space = add_prefix_space

        self.byte_encoder: dict[int, str] = self._byte_encoder()
        self.byte_decoder: dict[str, int] = {v: k for k, v in self.byte_encoder.items()}

        super().__init__(
            vocab,
            merges,
            vocab_file,
            merges_file,
            unk_token,
            pad_token,
            bos_token,
            eos_token,
            lowercase,
            clean_text,
            end_of_word_suffix,
        )
        if loaded_from_artifacts:
            self._validate_byte_vocab(self.vocab)

    @staticmethod
    @lru_cache(maxsize=1)
    def _byte_encoder() -> dict[int, str]:
        bs = list(range(33, 127)) + list(range(161, 173)) + list(range(174, 256))
        cs = bs[:]
        n = 0
        for b in range(256):
            if b not in bs:
                bs.append(b)
                cs.append(256 + n)
                n += 1

        return {b: chr(c) for b, c in zip(bs, cs)}

    @staticmethod
    def _validate_byte_vocab(vocab: dict[str, int]) -> None:
        if not vocab:
            return
        required = set(ByteBPETokenizer._byte_encoder().values())
        missing = required.difference(vocab.keys())
        if missing:
            raise ValueError(
                "Invalid ByteBPE vocabulary: missing required byte symbols "
                f"({len(missing)} missing)."
            )

    def _normalize_text(self, text: str) -> str:
        if self.clean_text:
            text = clean_text(text)
        if self.lowercase:
            text = text.lower()
        if self.add_prefix_space and text and not text.startswith(" "):
            text = " " + text
        return text

    def _bytes_to_symbols(self, text: str) -> list[str]:
        pieces = re.findall(r"\S+\s*|\s+", text)
        out: list[str] = []
        for p in pieces:
            if not p:
                continue
            b = p.encode("utf-8")
            out.append("".join(self.byte_encoder[x] for x in b))
        return out

    def _build_word_frequency(self, texts: Iterable[str]) -> Counter[str]:
        freq: Counter[str] = Counter()
        for text in texts:
            norm = self._normalize_text(text)
            for tok in self._bytes_to_symbols(norm):
                if tok:
                    freq[tok] += 1
        return freq

    def fit(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
        verbose: bool = False,
    ) -> ByteBPETokenizer:
        min_bbpe_vocab = len(self.all_special_tokens) + 256
        if vocab_size < min_bbpe_vocab:
            raise ValueError(
                "'vocab_size' must be >= number of special tokens + 256 for ByteBPE "
                f"(got {vocab_size}, need at least {min_bbpe_vocab})."
            )
        out = super().fit(
            texts=texts,
            vocab_size=vocab_size,
            min_frequency=min_frequency,
            verbose=verbose,
        )
        self._validate_byte_vocab(self.vocab)
        return out

    def tokenize(self, text: str) -> list[str]:
        norm = self._normalize_text(text)
        byte_tokens = self._bytes_to_symbols(norm)

        out: list[str] = []
        for tok in byte_tokens:
            out.extend(self._bpe_tokenize(tok))

        return out

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        bs = bytearray()
        special_tokens = set(self.all_special_tokens)
        for tok in tokens:
            if tok in special_tokens:
                continue

            piece = tok
            if self.end_of_word_suffix and piece.endswith(self.end_of_word_suffix):
                piece = piece[: -len(self.end_of_word_suffix)]

            for ch in piece:
                b = self.byte_decoder.get(ch)
                if b is not None:
                    bs.append(b)

        return bs.decode("utf-8", errors="replace")

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        paths = super().save_pretrained(save_directory)

        save_path = Path(save_directory)
        cfg = self._load_json(save_path / "tokenizer_config.json")
        cfg["add_prefix_space"] = self.add_prefix_space

        self._save_json(cfg, save_path / "tokenizer_config.json")
        return paths

    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path: Path | str, **kwargs: Any
    ) -> ByteBPETokenizer:
        path = Path(pretrained_model_name_or_path)
        base_dir = path if path.is_dir() else path.parent

        allowed = {
            "unk_token",
            "pad_token",
            "bos_token",
            "eos_token",
            "lowercase",
            "clean_text",
            "end_of_word_suffix",
            "add_prefix_space",
        }
        init_kwargs = cls._load_tokenizer_config(base_dir, allowed_keys=allowed)
        init_kwargs.update(kwargs)

        if path.is_dir():
            vocab_file = base_dir / "vocab.json"
            merges_file = base_dir / "merges.txt"
        else:
            vocab_file = path
            merges_file = base_dir / "merges.txt"

        return cls(vocab_file=vocab_file, merges_file=merges_file, **init_kwargs)


class ByteBPETokenizerFast(BPETokenizerFast):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        merges: list[tuple[str, str]] | None = None,
        vocab_file: Path | str | None = None,
        merges_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        bos_token: SpecialTokens | str | None = None,
        eos_token: SpecialTokens | str | None = None,
        lowercase: bool = False,
        clean_text: bool = True,
        add_prefix_space: bool = False,
        end_of_word_suffix: str = "",
    ) -> None:
        loaded_from_artifacts = (
            vocab is not None
            or merges is not None
            or vocab_file is not None
            or merges_file is not None
        )
        if vocab is not None and vocab_file is not None:
            raise ValueError("Provide only one of 'vocab' or 'vocab_file'.")
        if merges is not None and merges_file is not None:
            raise ValueError("Provide only one of 'merges' or 'merges_file'.")

        Tokenizer.__init__(self, unk_token, pad_token, bos_token, eos_token)

        backend_vocab_file: Path | None = None
        backend_merges_file: Path | None = None
        if vocab_file is not None:
            vocab = BPETokenizer._load_vocab(vocab_file)
        else:
            backend_vocab_file = None
        if merges_file is not None:
            merges = BPETokenizer._load_merges(merges_file)
        else:
            backend_merges_file = None

        self.lowercase = lowercase
        self.clean_text = clean_text
        self.add_prefix_space = add_prefix_space
        self.end_of_word_suffix = end_of_word_suffix

        self.vocab: dict[str, int] = dict(vocab or {})
        self.merges: list[tuple[str, str]] = list(merges or [])

        self._backend = _C_ByteBPETokenizer(
            vocab=vocab,
            merges=merges,
            vocab_file=backend_vocab_file,
            merges_file=backend_merges_file,
            unk_token=self.unk_token,
            pad_token=self.pad_token,
            bos_token=self.bos_token,
            eos_token=self.eos_token,
            lowercase=self.lowercase,
            clean_text=self.clean_text,
            add_prefix_space=self.add_prefix_space,
            end_of_word_suffix=self.end_of_word_suffix,
        )
        self._sync_state_from_backend()
        if loaded_from_artifacts:
            ByteBPETokenizer._validate_byte_vocab(self.vocab)

    def fit(
        self,
        texts: Iterable[str],
        vocab_size: int,
        min_frequency: int = 2,
    ) -> ByteBPETokenizerFast:
        min_bbpe_vocab = len(self.all_special_tokens) + 256
        if vocab_size < min_bbpe_vocab:
            raise ValueError(
                "'vocab_size' must be >= number of special tokens + 256 for ByteBPE "
                f"(got {vocab_size}, need at least {min_bbpe_vocab})."
            )
        out = super().fit(
            texts=texts,
            vocab_size=vocab_size,
            min_frequency=min_frequency,
        )
        ByteBPETokenizer._validate_byte_vocab(self.vocab)
        return out

    def save_pretrained(self, save_directory: Path | str) -> list[str]:
        paths = super().save_pretrained(save_directory)
        save_path = Path(save_directory)

        cfg = self._load_json(save_path / "tokenizer_config.json")
        cfg["add_prefix_space"] = self.add_prefix_space

        self._save_json(cfg, save_path / "tokenizer_config.json")
        return paths

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: str | Path,
        **kwargs: Any,
    ) -> "ByteBPETokenizerFast":
        path = Path(pretrained_model_name_or_path)
        if path.is_dir():
            base_dir = path
            vocab_file = base_dir / "vocab.json"
            merges_file = base_dir / "merges.txt"
        else:
            base_dir = path.parent
            vocab_file = path

            if vocab_file.name != "vocab.json":
                raise ValueError(
                    "Expected a directory containing 'vocab.json'/'merges.txt' "
                    "or a direct path to 'vocab.json'."
                )
            merges_file = base_dir / "merges.txt"

        if not vocab_file.exists():
            raise FileNotFoundError(f"Cannot find vocabulary file: {vocab_file}")
        if not merges_file.exists():
            raise FileNotFoundError(f"Cannot find merges file: {merges_file}")

        allowed = {
            "unk_token",
            "pad_token",
            "bos_token",
            "eos_token",
            "lowercase",
            "clean_text",
            "add_prefix_space",
            "end_of_word_suffix",
        }
        init_kwargs = cls._load_tokenizer_config(base_dir, allowed_keys=allowed)
        init_kwargs.update(kwargs)

        return cls(vocab_file=vocab_file, merges_file=merges_file, **init_kwargs)



from dataclasses import dataclass, field
from typing import Any, ClassVar

import lucid.nn as nn
import lucid.nn.functional as F

from lucid import register_model
from lucid._tensor import Tensor

from .resnet import ResNet, ResNetConfig

__all__ = [
    "ResNeSt",
    "ResNeStConfig",
    "resnest_14",
    "resnest_26",
    "resnest_50",
    "resnest_101",
    "resnest_200",
    "resnest_269",
    "resnest_50_4s2x40d",
    "resnest_50_1s4x24d",
]


@dataclass
class ResNeStConfig:
    layers: tuple[int, int, int, int] | list[int]
    base_width: int = 64
    stem_width: int = 32
    cardinality: int = 1
    radix: int = 2
    avd: bool = True
    num_classes: int = 1000
    in_channels: int = 3
    avg_down: bool = False
    channels: tuple[int, int, int, int] | list[int] = (64, 128, 256, 512)
    block_args: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.layers = tuple(self.layers)
        self.channels = tuple(self.channels)
        if not isinstance(self.block_args, dict):
            raise TypeError("block_args must be a dictionary")
        self.block_args = dict(self.block_args)

        if len(self.layers) != 4:
            raise ValueError("layers must contain exactly 4 stage depths")
        if any(not isinstance(depth, int) or depth <= 0 for depth in self.layers):
            raise ValueError("layers values must be positive integers")
        if self.base_width <= 0:
            raise ValueError("base_width must be greater than 0")
        if self.stem_width <= 0:
            raise ValueError("stem_width must be greater than 0")
        if self.cardinality <= 0:
            raise ValueError("cardinality must be greater than 0")
        if self.radix < 0:
            raise ValueError("radix must be greater than or equal to 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if len(self.channels) != 4:
            raise ValueError("channels must contain exactly 4 stage widths")
        if any(
            not isinstance(channel, int) or channel <= 0 for channel in self.channels
        ):
            raise ValueError("channels values must be positive integers")


class _RadixSoftmax(nn.Module):
    def __init__(self, radix: int, cardinality: int) -> None:
        super().__init__()
        self.radix = radix
        self.cardinality = cardinality

    def forward(self, x: Tensor) -> Tensor:
        N = x.shape[0]
        if self.radix > 1:
            x = x.reshape(N, self.radix, self.cardinality, -1)
            x = F.softmax(x, axis=1)
            x = x.reshape(N, -1)
        else:
            x = F.sigmoid(x)

        return x


class _SplitAttention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int | None = None,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int | str = 0,
        groups: int = 1,
        bias: bool = False,
        radix: int = 2,
        reduce: int = 4,
        drop_layer: nn.Module | None = None,
    ) -> None:
        super().__init__()
        out_channels = out_channels or in_channels
        mid_channels = out_channels * radix
        attn_channels = max(32, int(in_channels * radix / reduce))

        self.conv = nn.Conv2d(
            in_channels,
            mid_channels,
            kernel_size,
            stride,
            padding,
            groups=groups * radix,
            bias=bias,
        )
        self.bn0 = nn.BatchNorm2d(mid_channels)
        self.drop = drop_layer() if drop_layer is not None else nn.Identity()
        self.relu0 = nn.ReLU()

        self.fc1 = nn.ConvBNReLU2d(
            out_channels, attn_channels, kernel_size=1, groups=groups
        )
        self.fc2 = nn.Conv2d(attn_channels, mid_channels, kernel_size=1, groups=groups)
        self.rsoftmax = _RadixSoftmax(radix, cardinality=groups)
        self.radix = radix

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.bn0(x)
        x = self.drop(x)
        x = self.relu0(x)

        N, C, H, W = x.shape
        if self.radix > 1:
            x = x.reshape(N, self.radix, C // self.radix, H, W)
            x_gap = x.sum(axis=1)
        else:
            x_gap = x

        x_gap = x_gap.mean(axis=(2, 3), keepdims=True)
        x_gap = self.fc1(x_gap)
        x_attn = self.fc2(x_gap)
        x_attn = self.rsoftmax(x_attn).reshape(N, -1, 1, 1)

        if self.radix > 1:
            out = (x * x_attn.reshape(N, self.radix, C // self.radix, 1, 1)).sum(axis=1)
        else:
            out = x * x_attn
        return out


class _ResNeStBottleneck(nn.Module):
    expansion: ClassVar[int] = 4

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        radix: int = 1,
        cardinality: int = 1,
        base_width: int = 64,
        avd: bool = False,
        is_first: bool = False,
    ) -> None:
        super().__init__()
        group_width = int(out_channels * (base_width / 64.0)) * cardinality
        if avd and (stride > 1 or is_first):
            avd_stride = stride
            stride = 1
        else:
            avd_stride = 0

        self.conv1 = nn.ConvBNReLU2d(
            in_channels, group_width, kernel_size=1, conv_bias=False
        )

        conv2_args = (group_width, group_width, 3, stride, 1)
        if radix >= 1:
            self.conv2 = _SplitAttention(*conv2_args, groups=cardinality, radix=radix)
        else:
            self.conv2 = nn.ConvBNReLU2d(
                *conv2_args, groups=cardinality, conv_bias=False
            )

        self.avd_last = (
            nn.AvgPool2d(kernel_size=3, stride=avd_stride, padding=1)
            if avd_stride > 0
            else None
        )
        self.conv3 = nn.Conv2d(
            group_width, out_channels * self.expansion, kernel_size=1, bias=False
        )
        self.bn3 = nn.BatchNorm2d(out_channels * self.expansion)
        self.relu = nn.ReLU()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        shortcut = x

        out = self.conv1(x)
        out = self.conv2(out)

        if self.avd_last is not None:
            out = self.avd_last(out)

        out = self.bn3(self.conv3(out))
        if self.downsample is not None:
            shortcut = self.downsample(x)

        out += shortcut
        out = self.relu(out)
        return out


class ResNeSt(ResNet):
    def __init__(self, config: ResNeStConfig) -> None:
        block_args = {
            **config.block_args,
            "base_width": config.base_width,
            "cardinality": config.cardinality,
            "radix": config.radix,
            "avd": config.avd,
        }
        super().__init__(
            ResNetConfig(
                block=_ResNeStBottleneck,
                layers=config.layers,
                num_classes=config.num_classes,
                in_channels=config.in_channels,
                stem_width=config.stem_width,
                stem_type="deep",
                avg_down=config.avg_down,
                channels=config.channels,
                block_args=block_args,
            )
        )
        self.config = config
        self.base_width = config.base_width
        self.stem_width = config.stem_width
        self.cardinality = config.cardinality
        self.radix = config.radix
        self.avd = config.avd


def _build_resnest_config(
    *,
    layers: tuple[int, int, int, int] | list[int],
    base_width: int,
    stem_width: int,
    cardinality: int,
    radix: int,
    avd: bool,
    num_classes: int,
    kwargs: dict[str, Any] | None = None,
) -> ResNeStConfig:
    kwargs = {} if kwargs is None else dict(kwargs)
    locked_fields = {
        "layers",
        "base_width",
        "stem_width",
        "cardinality",
        "radix",
        "avd",
    }
    if locked_fields & kwargs.keys():
        raise TypeError(
            "factory variants do not allow overriding preset layers, base_width, "
            "stem_width, cardinality, radix, or avd"
        )

    return ResNeStConfig(
        layers=layers,
        base_width=base_width,
        stem_width=stem_width,
        cardinality=cardinality,
        radix=radix,
        avd=avd,
        num_classes=num_classes,
        **kwargs,
    )


@register_model
def resnest_14(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [1, 1, 1, 1]
    config = _build_resnest_config(
        layers=layers,
        base_width=64,
        stem_width=32,
        cardinality=1,
        radix=2,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_26(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [2, 2, 2, 2]
    config = _build_resnest_config(
        layers=layers,
        base_width=64,
        stem_width=32,
        cardinality=1,
        radix=2,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_50(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [3, 4, 6, 3]
    config = _build_resnest_config(
        layers=layers,
        base_width=64,
        stem_width=32,
        cardinality=1,
        radix=2,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_101(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [3, 4, 23, 3]
    config = _build_resnest_config(
        layers=layers,
        base_width=64,
        stem_width=64,
        cardinality=1,
        radix=2,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_200(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [3, 24, 36, 3]
    config = _build_resnest_config(
        layers=layers,
        base_width=64,
        stem_width=64,
        cardinality=1,
        radix=2,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_269(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [3, 30, 48, 8]
    config = _build_resnest_config(
        layers=layers,
        base_width=64,
        stem_width=64,
        cardinality=1,
        radix=2,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_50_4s2x40d(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [3, 4, 6, 3]
    config = _build_resnest_config(
        layers=layers,
        base_width=40,
        stem_width=32,
        cardinality=2,
        radix=4,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)


@register_model
def resnest_50_1s4x24d(num_classes: int = 1000, **kwargs) -> ResNeSt:
    layers = [3, 4, 6, 3]
    config = _build_resnest_config(
        layers=layers,
        base_width=24,
        stem_width=32,
        cardinality=4,
        radix=1,
        avd=True,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return ResNeSt(config)



from typing import Any, Callable, ClassVar

import numpy as np
import torch

import lucid
from lucid.test._utils.test_utils import _unwrap_callable
from lucid.test.core import TensorOpTorchCase, TensorOpWithTorchBase


class _NNFunctionalTorchBase(TensorOpWithTorchBase):
    _collect_nested_test_classes = True
    case_name: ClassVar[str]
    inputs: ClassVar[tuple[Any, ...]]

    forward_op: ClassVar[Callable[..., Any]]
    torch_forward: ClassVar[Callable[..., Any]]
    expected_forward: ClassVar[Any | None] = None

    requires_grad: ClassVar[bool | tuple[bool, ...]] = True
    input_dtypes: ClassVar[tuple[Any | None, ...] | None] = None

    expected_input_grads: ClassVar[tuple[object | None, ...] | None] = None
    compare_grad_with_torch: ClassVar[bool] = True
    check_finite: ClassVar[bool] = True

    rtol: ClassVar[float] = 1e-7
    atol: ClassVar[float] = 1e-8

    def tensor_op_devices(self) -> tuple[str, ...]:
        return ("cpu",)

    def _normalize_bool_flags(
        self,
        flags: bool | tuple[bool, ...],
        expected_len: int,
    ) -> tuple[bool, ...]:
        if isinstance(flags, bool):
            return (bool(flags),) * expected_len
        if len(flags) == 1 and expected_len > 1:
            return (bool(flags[0]),) * expected_len
        if len(flags) != expected_len:
            raise AssertionError(
                f"requires_grad length {len(flags)} does not match inputs "
                f"(expected {expected_len}) for case '{self.case_name}'."
            )
        return tuple(bool(v) for v in flags)

    def _normalize_input_dtypes(
        self,
        dtypes: tuple[Any | None, ...] | None,
        expected_len: int,
    ) -> tuple[Any | None, ...]:
        if dtypes is None:
            return (None,) * expected_len
        if len(dtypes) == 1 and expected_len > 1:
            return (dtypes[0],) * expected_len
        if len(dtypes) != expected_len:
            raise AssertionError(
                f"input_dtypes length {len(dtypes)} does not match inputs "
                f"(expected {expected_len}) for case '{self.case_name}'."
            )
        return tuple(dtypes)

    def _build_inputs(self) -> tuple[Any, ...]:
        requires = self._normalize_bool_flags(self.requires_grad, len(self.inputs))
        dtypes = self._normalize_input_dtypes(self.input_dtypes, len(self.inputs))
        outputs: list[Any] = []
        for item, requires_grad, dtype in zip(self.inputs, requires, dtypes):
            if item is None:
                outputs.append(None)
                continue
            outputs.append(self.tensor(item, requires_grad=requires_grad, dtype=dtype))
        return tuple(outputs)

    def _to_torch_inputs(self, inputs: tuple[Any, ...], torch: Any) -> tuple[Any, ...]:
        output: list[Any] = []
        for item in inputs:
            if item is None:
                output.append(None)
                continue
            output.extend(super()._to_torch_inputs((item,), torch))
        return tuple(output)

    def _to_numpy(self, value: Any) -> Any:
        if isinstance(value, tuple):
            return tuple(self._to_numpy(item) for item in value)
        if isinstance(value, torch.Tensor):
            return value.detach().cpu().numpy()
        if hasattr(value, "data"):
            return np.array(value.data)
        return np.array(value)

    def _build_expected_forward(self, torch_forward: Callable[..., Any]) -> Any:
        torch_inputs = self._to_torch_inputs(self._build_inputs(), torch)
        torch_out = torch_forward(*torch_inputs)
        return self._to_numpy(torch_out)

    def tensor_op_cases(self) -> tuple[TensorOpTorchCase, ...]:
        forward_op = _unwrap_callable(self.forward_op)
        torch_forward = _unwrap_callable(self.torch_forward)

        return (
            TensorOpTorchCase(
                name=self.case_name,
                build_inputs=lambda _: self._build_inputs(),
                forward_op=forward_op,
                expected_forward=(
                    self.expected_forward
                    if self.expected_forward is not None
                    else self._build_expected_forward(torch_forward)
                ),
                expected_input_grads=self.expected_input_grads,
                requires_grad=self._normalize_bool_flags(
                    self.requires_grad, len(self.inputs)
                ),
                check_finite=self.check_finite,
                torch_forward=torch_forward,
                compare_grad_with_torch=self.compare_grad_with_torch,
                rtol=self.rtol,
                atol=self.atol,
            ),
        )


class TestActivation(_NNFunctionalTorchBase):
    __test__ = False

    class TestRelu(_NNFunctionalTorchBase):
        case_name = "relu"
        inputs = (np.array([[-1.0, 2.0], [3.0, -4.0]], dtype=np.float64),)
        forward_op = staticmethod(lucid.nn.functional.relu)
        torch_forward = staticmethod(torch.relu)

    class TestLeakyRelu(_NNFunctionalTorchBase):
        case_name = "leaky_relu"
        inputs = (np.array([[-1.0, 2.0], [-3.0, 4.0]], dtype=np.float64),)
        forward_op = staticmethod(
            lambda a: lucid.nn.functional.leaky_relu(a, negative_slope=0.2)
        )
        torch_forward = staticmethod(
            lambda a: torch.nn.functional.leaky_relu(a, negative_slope=0.2)
        )

    class TestElu(_NNFunctionalTorchBase):
        case_name = "elu"
        inputs = (np.array([[-1.0, 2.0], [-3.0, 4.0]], dtype=np.float64),)
        forward_op = staticmethod(lambda a: lucid.nn.functional.elu(a, alpha=1.2))
        torch_forward = staticmethod(lambda a: torch.nn.functional.elu(a, alpha=1.2))

    class TestSelu(_NNFunctionalTorchBase):
        case_name = "selu"
        inputs = (np.array([[-1.0, 2.0], [-3.0, 4.0]], dtype=np.float64),)
        forward_op = staticmethod(lucid.nn.functional.selu)
        torch_forward = staticmethod(torch.selu)

    class TestSilu(_NNFunctionalTorchBase):
        case_name = "silu"
        inputs = (np.array([[-1.0, 2.0], [-3.0, 4.0]], dtype=np.float64),)
        forward_op = staticmethod(lucid.nn.functional.silu)
        torch_forward = staticmethod(torch.nn.functional.silu)

    class TestSoftmax(_NNFunctionalTorchBase):
        case_name = "softmax"
        inputs = (np.array([[1.0, 2.0, 3.0], [1.0, 1.0, 1.0]], dtype=np.float64),)
        forward_op = staticmethod(lambda a: lucid.nn.functional.softmax(a, axis=1))
        torch_forward = staticmethod(lambda a: torch.nn.functional.softmax(a, dim=1))


class TestLinear(_NNFunctionalTorchBase):
    __test__ = False

    class TestLinear(_NNFunctionalTorchBase):
        case_name = "linear"
        inputs = (
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([[2.0, 3.0], [4.0, 5.0]], dtype=np.float64),
            np.array([1.0, 1.0], dtype=np.float64),
        )
        forward_op = staticmethod(lucid.nn.functional.linear)
        torch_forward = staticmethod(torch.nn.functional.linear)

    class TestLinearNoBias(_NNFunctionalTorchBase):
        case_name = "linear_no_bias"
        inputs = (
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([[2.0, 3.0], [4.0, 5.0]], dtype=np.float64),
        )
        forward_op = staticmethod(lucid.nn.functional.linear)
        torch_forward = staticmethod(torch.nn.functional.linear)

    class TestBilinear(_NNFunctionalTorchBase):
        case_name = "bilinear"
        inputs = (
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([[5.0, 6.0], [7.0, 8.0]], dtype=np.float64),
            np.array([[[1.0, 2.0], [3.0, 4.0]]], dtype=np.float64),
            np.array([1.0], dtype=np.float64),
        )
        forward_op = staticmethod(lucid.nn.functional.bilinear)
        torch_forward = staticmethod(torch.bilinear)


class TestLoss(_NNFunctionalTorchBase):
    __test__ = False

    class TestMSEMean(_NNFunctionalTorchBase):
        case_name = "mse_mean"
        inputs = (
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([[1.5, 1.0], [2.5, 3.5]], dtype=np.float64),
        )
        forward_op = staticmethod(lambda a, b: lucid.nn.functional.mse_loss(a, b))
        torch_forward = staticmethod(
            lambda a, b: torch.nn.functional.mse_loss(a, b, reduction="mean")
        )
        requires_grad = (True, False)

    class TestMSESum(_NNFunctionalTorchBase):
        case_name = "mse_sum"
        inputs = (
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([[2.0, 1.0], [0.0, 4.0]], dtype=np.float64),
        )
        forward_op = staticmethod(
            lambda a, b: lucid.nn.functional.mse_loss(a, b, reduction="sum")
        )
        torch_forward = staticmethod(
            lambda a, b: torch.nn.functional.mse_loss(a, b, reduction="sum")
        )
        requires_grad = (True, False)

    class TestHuber(_NNFunctionalTorchBase):
        case_name = "huber"
        inputs = (
            np.array([[1.0, 2.0], [3.0, 4.0]], dtype=np.float64),
            np.array([[1.2, 1.9], [2.0, 5.0]], dtype=np.float64),
        )
        forward_op = staticmethod(
            lambda a, b: lucid.nn.functional.huber_loss(a, b, delta=1.0)
        )
        torch_forward = staticmethod(
            lambda a, b: torch.nn.functional.smooth_l1_loss(
                a, b, beta=1.0, reduction="mean"
            )
        )
        requires_grad = (True, False)
        atol = 1e-6
        rtol = 1e-6

    class TestBinaryCrossEntropy(_NNFunctionalTorchBase):
        case_name = "bce"
        inputs = (
            np.array([[0.1, 0.7], [0.2, 0.9]], dtype=np.float64),
            np.array([[0.0, 1.0], [1.0, 0.0]], dtype=np.float64),
        )
        forward_op = staticmethod(
            lambda a, b: lucid.nn.functional.binary_cross_entropy(a, b)
        )
        torch_forward = staticmethod(torch.nn.functional.binary_cross_entropy)
        requires_grad = (True, False)

    class TestBCEWithLogits(_NNFunctionalTorchBase):
        case_name = "bce_with_logits"
        inputs = (
            np.array([[0.2, -1.0], [1.5, 0.1]], dtype=np.float64),
            np.array([[1.0, 0.0], [0.0, 1.0]], dtype=np.float64),
        )
        forward_op = staticmethod(
            lambda a, b: lucid.nn.functional.binary_cross_entropy_with_logits(
                a, b, pos_weight=lucid.ones_like(a)
            )
        )
        torch_forward = staticmethod(
            lambda a, b: torch.nn.functional.binary_cross_entropy_with_logits(
                a, b, pos_weight=torch.ones_like(a)
            )
        )
        requires_grad = (True, False)

    class TestCrossEntropy(_NNFunctionalTorchBase):
        case_name = "cross_entropy"
        inputs = (
            np.array([[2.0, 1.0, 0.1], [0.1, 0.2, 3.0]], dtype=np.float64),
            np.array([1, 2], dtype=np.int64),
        )
        forward_op = staticmethod(
            lambda a, b: lucid.nn.functional.cross_entropy(a, b, reduction="mean")
        )
        torch_forward = staticmethod(
            lambda a, b: torch.nn.functional.cross_entropy(a, b, reduction="mean")
        )
        requires_grad = (True, False)


class TestNormalization(_NNFunctionalTorchBase):
    __test__ = False

    class TestNormalize(_NNFunctionalTorchBase):
        case_name = "normalize"
        inputs = (np.array([[3.0, 4.0], [1.0, 2.0]], dtype=np.float64),)
        forward_op = staticmethod(
            lambda a: lucid.nn.functional.normalize(a, ord=2, axis=1)
        )
        torch_forward = staticmethod(
            lambda a: torch.nn.functional.normalize(a, p=2, dim=1, eps=1e-12)
        )

    class TestLayerNorm(_NNFunctionalTorchBase):
        case_name = "layer_norm"
        inputs = (
            np.array(
                [[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]],
                dtype=np.float64,
            ),
        )
        forward_op = staticmethod(
            lambda a: lucid.nn.functional.layer_norm(a, normalized_shape=(4,))
        )
        torch_forward = staticmethod(
            lambda a: torch.nn.functional.layer_norm(
                a, normalized_shape=(4,), weight=None, bias=None
            )
        )

    class TestBatchNorm(_NNFunctionalTorchBase):
        case_name = "batch_norm"
        inputs = (
            np.array(
                [[1.0, 2.0, 3.0, 4.0], [4.0, 5.0, 6.0, 7.0], [2.0, 1.0, 0.0, 2.0]],
                dtype=np.float64,
            ).reshape(1, 3, 2, 2),
            np.array([0.5, 0.4, 0.3], dtype=np.float64),
            np.array([1.5, 1.2, 1.1], dtype=np.float64),
            np.array([1.0, 1.0, 1.0], dtype=np.float64),
            np.array([0.1, 0.0, -0.1], dtype=np.float64),
        )
        forward_op = staticmethod(
            lambda a, b, c, d, e: lucid.nn.functional.batch_norm(
                a, b, c, d, e, training=False, momentum=0.1, eps=1e-5
            )
        )
        torch_forward = staticmethod(
            lambda a, b, c, d, e: torch.nn.functional.batch_norm(
                a, b, c, d, e, training=False, momentum=0.1, eps=1e-5
            )
        )
        requires_grad = (True, False, False, True, True)

    class TestGroupNorm(_NNFunctionalTorchBase):
        case_name = "group_norm"
        inputs = (
            np.arange(1, 33, dtype=np.float64).reshape(2, 4, 2, 2),
            np.array([1.0, 1.0, 1.0, 1.0], dtype=np.float64),
            np.array([0.0, 0.0, 0.0, 0.0], dtype=np.float64),
        )
        forward_op = staticmethod(
            lambda a, c, d: lucid.nn.functional.group_norm(a, 2, c, d, eps=1e-5)
        )
        torch_forward = staticmethod(
            lambda a, c, d: torch.nn.functional.group_norm(
                a, num_groups=2, weight=c, bias=d, eps=1e-5
            )
        )
        requires_grad = (True, True, True)


class TestPool(_NNFunctionalTorchBase):
    __test__ = False

    class TestAvgPool2d(_NNFunctionalTorchBase):
        case_name = "avg_pool2d"
        inputs = (np.arange(1, 17, dtype=np.float64).reshape(1, 1, 4, 4),)
        forward_op = staticmethod(
            lambda a: lucid.nn.functional.avg_pool2d(
                a, kernel_size=(2, 2), stride=(2, 2)
            )
        )
        torch_forward = staticmethod(
            lambda a: torch.nn.functional.avg_pool2d(
                a, kernel_size=(2, 2), stride=(2, 2)
            )
        )

    class TestMaxPool2d(_NNFunctionalTorchBase):
        case_name = "max_pool2d"
        inputs = (np.arange(1, 17, dtype=np.float64).reshape(1, 1, 4, 4),)
        forward_op = staticmethod(
            lambda a: lucid.nn.functional.max_pool2d(
                a, kernel_size=(2, 2), stride=(2, 2)
            )
        )
        torch_forward = staticmethod(
            lambda a: torch.nn.functional.max_pool2d(
                a, kernel_size=(2, 2), stride=(2, 2)
            )
        )


class TestConvolution(_NNFunctionalTorchBase):
    __test__ = False

    class TestConv2d(_NNFunctionalTorchBase):
        case_name = "conv2d"
        inputs = (
            np.arange(16.0, dtype=np.float64).reshape(1, 1, 4, 4),
            np.array(
                [[[[1.0, 0.0, -1.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 1.0]]]],
                dtype=np.float64,
            ),
            np.array([0.25], dtype=np.float64),
        )
        forward_op = staticmethod(lucid.nn.functional.conv2d)
        torch_forward = staticmethod(
            lambda a, b, c: torch.nn.functional.conv2d(
                a, b, c, stride=1, padding=0, dilation=1
            )
        )


class TestEmbedding(_NNFunctionalTorchBase):
    __test__ = False

    class TestEmbedding(_NNFunctionalTorchBase):
        case_name = "embedding"
        inputs = (
            np.array([[0, 1, 2], [1, 3, 4]], dtype=np.int64),
            np.array(
                [
                    [1.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0],
                    [0.0, 0.0, 1.0],
                    [1.0, 1.0, 1.0],
                    [2.0, 2.0, 2.0],
                ],
                dtype=np.float64,
            ),
        )
        forward_op = staticmethod(
            lambda a, b: lucid.nn.functional.embedding(
                a, b, padding_idx=0, max_norm=3.0, norm_type=2.0
            )
        )
        torch_forward = staticmethod(
            lambda a, b: torch.nn.functional.embedding(
                a, b, padding_idx=0, max_norm=3.0, norm_type=2.0
            )
        )
        requires_grad = (False, True)

    class TestOneHot(_NNFunctionalTorchBase):
        case_name = "one_hot"
        inputs = (np.array([[0, 1, 2], [2, 3, 0]], dtype=np.int64),)
        forward_op = staticmethod(
            lambda a: lucid.nn.functional.one_hot(a, num_classes=5)
        )
        torch_forward = staticmethod(
            lambda a: torch.nn.functional.one_hot(a, num_classes=5)
        )
        requires_grad = (False,)
        compare_grad_with_torch = False



from dataclasses import dataclass
from typing import Any

from lucid._jit.executor import CompiledPlan


@dataclass(frozen=True)
class CacheKey:
    shape_dtypes: tuple[tuple[tuple[int, ...], Any, str], ...]
    grad_enabled: bool
    training_mode: bool

    @classmethod
    def from_inputs(
        cls,
        inputs: tuple,
        grad_enabled: bool,
        training_mode: bool,
    ) -> CacheKey:
        shape_dtypes = tuple(
            (tuple(t.shape), t.dtype, t.device) for t in inputs if hasattr(t, "shape")
        )
        return cls(
            shape_dtypes=shape_dtypes,
            grad_enabled=grad_enabled,
            training_mode=training_mode,
        )


class PlanCache:
    def __init__(self, max_entries: int = 8) -> None:
        self._cache: dict[CacheKey, CompiledPlan] = {}
        self._max_entries = max_entries

    def get(self, key: CacheKey) -> CompiledPlan | None:
        return self._cache.get(key)

    def put(self, key: CacheKey, plan: CompiledPlan) -> None:
        if len(self._cache) >= self._max_entries:
            oldest = next(iter(self._cache))
            del self._cache[oldest]
        self._cache[key] = plan

    def invalidate(self) -> None:
        self._cache.clear()

    def __len__(self) -> int:
        return len(self._cache)



from dataclasses import dataclass

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin

__all__ = [
    "DenseNet",
    "DenseNetConfig",
    "densenet_121",
    "densenet_169",
    "densenet_201",
    "densenet_264",
]


@dataclass
class DenseNetConfig:
    block_config: tuple[int, int, int, int] | list[int]
    growth_rate: int = 32
    num_init_features: int = 64
    num_classes: int = 1000
    in_channels: int = 3
    bottleneck: int = 4
    compression: float = 0.5

    def __post_init__(self) -> None:
        self.block_config = tuple(self.block_config)

        if len(self.block_config) != 4:
            raise ValueError("block_config must contain exactly 4 dense block depths")
        if any(not isinstance(depth, int) or depth <= 0 for depth in self.block_config):
            raise ValueError("block_config values must be positive integers")
        if self.growth_rate <= 0:
            raise ValueError("growth_rate must be greater than 0")
        if self.num_init_features <= 0:
            raise ValueError("num_init_features must be greater than 0")
        if self.num_classes <= 0:
            raise ValueError("num_classes must be greater than 0")
        if self.in_channels <= 0:
            raise ValueError("in_channels must be greater than 0")
        if self.bottleneck <= 0:
            raise ValueError("bottleneck must be greater than 0")
        if self.compression <= 0 or self.compression > 1:
            raise ValueError("compression must be in the range (0, 1]")


class DenseNet(nn.Module, PreTrainedModelMixin):
    def __init__(self, config: DenseNetConfig) -> None:
        super().__init__()
        self.config = config
        self.growth_rate = config.growth_rate
        self.num_init_features = config.num_init_features
        self.bottleneck = config.bottleneck
        self.compression = config.compression

        self.conv0 = nn.ConvBNReLU2d(
            config.in_channels,
            config.num_init_features,
            kernel_size=7,
            stride=2,
            padding=3,
            conv_bias=False,
        )
        self.pool0 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.blocks = nn.ModuleList()
        self.transitions = nn.ModuleList()

        in_channels = config.num_init_features
        for i, num_layers in enumerate(config.block_config):
            block = _DenseBlock(
                in_channels,
                num_layers,
                config.growth_rate,
                bottleneck=config.bottleneck,
            )
            self.blocks.append(block)

            in_channels += num_layers * config.growth_rate

            if i != len(config.block_config) - 1:
                out_channels = int(in_channels * config.compression)
                transition = _TransitionLayer(in_channels, out_channels)
                self.transitions.append(transition)

                in_channels = out_channels

        self.bn_final = nn.BatchNorm2d(in_channels)
        self.relu = nn.ReLU()

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(in_channels, config.num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.pool0(self.conv0(x))

        for i, block in enumerate(self.blocks):
            x = block(x)
            if i < len(self.transitions):
                x = self.transitions[i](x)

        x = self.avgpool(self.relu(self.bn_final(x)))
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


class _DenseLayer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        growth_rate: int,
        bottleneck: int = 4,
    ) -> None:
        super().__init__()
        self.bn1 = nn.BatchNorm2d(in_channels)
        self.relu1 = nn.ReLU()
        self.conv1 = nn.Conv2d(
            in_channels, bottleneck * growth_rate, kernel_size=1, bias=False
        )

        self.bn2 = nn.BatchNorm2d(bottleneck * growth_rate)
        self.relu2 = nn.ReLU()
        self.conv2 = nn.Conv2d(
            bottleneck * growth_rate,
            growth_rate,
            kernel_size=3,
            padding=1,
            bias=False,
        )

    def forward(self, x: Tensor) -> Tensor:
        out = self.conv1(self.relu1(self.bn1(x)))
        out = self.conv2(self.relu2(self.bn2(out)))

        return lucid.concatenate([x, out], axis=1)


class _DenseBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_layers: int,
        growth_rate: int,
        bottleneck: int = 4,
    ) -> None:
        super().__init__()
        layers = [
            _DenseLayer(in_channels + i * growth_rate, growth_rate, bottleneck)
            for i in range(num_layers)
        ]
        self.layers = nn.ModuleList(layers)

    def forward(self, x: Tensor) -> Tensor:
        for layer in self.layers:
            x = layer(x)
        return x


class _TransitionLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.bn = nn.BatchNorm2d(in_channels)
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.pool = nn.AvgPool2d(kernel_size=2, stride=2)
        self.relu = nn.ReLU()

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(self.relu(self.bn(x)))
        x = self.pool(x)
        return x


def _build_densenet_config(
    *,
    block_config: tuple[int, int, int, int] | list[int],
    growth_rate: int,
    num_init_features: int,
    num_classes: int,
    kwargs: dict[str, object] | None = None,
) -> DenseNetConfig:
    kwargs = {} if kwargs is None else dict(kwargs)
    locked_fields = {"block_config", "growth_rate", "num_init_features"}
    if locked_fields & kwargs.keys():
        raise TypeError(
            "factory variants do not allow overriding preset block_config, growth_rate, or num_init_features"
        )

    return DenseNetConfig(
        block_config=block_config,
        growth_rate=growth_rate,
        num_init_features=num_init_features,
        num_classes=num_classes,
        **kwargs,
    )


@register_model
def densenet_121(num_classes: int = 1000, **kwargs) -> DenseNet:
    config = _build_densenet_config(
        block_config=(6, 12, 24, 16),
        growth_rate=32,
        num_init_features=64,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return DenseNet(config)


@register_model
def densenet_169(num_classes: int = 1000, **kwargs) -> DenseNet:
    config = _build_densenet_config(
        block_config=(6, 12, 32, 32),
        growth_rate=32,
        num_init_features=64,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return DenseNet(config)


@register_model
def densenet_201(num_classes: int = 1000, **kwargs) -> DenseNet:
    config = _build_densenet_config(
        block_config=(6, 12, 48, 32),
        growth_rate=32,
        num_init_features=64,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return DenseNet(config)


@register_model
def densenet_264(num_classes: int = 1000, **kwargs) -> DenseNet:
    config = _build_densenet_config(
        block_config=(6, 12, 64, 48),
        growth_rate=32,
        num_init_features=64,
        num_classes=num_classes,
        kwargs=kwargs,
    )
    return DenseNet(config)



from dataclasses import dataclass

import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor
from lucid.models.base import PreTrainedModelMixin

__all__ = ["AlexNet", "AlexNetConfig", "alexnet"]


@dataclass
class AlexNetConfig:
    num_classes: int = 1000
    in_channels: int = 3
    dropout: float = 0.5
    classifier_hidden_features: tuple[int, int] = (4096, 4096)

    def __post_init__(self) -> None:
        if self.num_classes <= 0:
            raise ValueError("AlexNet num_classes must be greater than 0.")

        if self.in_channels <= 0:
            raise ValueError("AlexNet in_channels must be greater than 0.")

        if not 0.0 <= self.dropout < 1.0:
            raise ValueError("AlexNet dropout must be in the range [0.0, 1.0).")

        if len(self.classifier_hidden_features) != 2:
            raise ValueError(
                "AlexNet classifier_hidden_features must contain exactly 2 values."
            )

        if any(hidden_dim <= 0 for hidden_dim in self.classifier_hidden_features):
            raise ValueError(
                "AlexNet classifier_hidden_features values must be greater than 0."
            )


class AlexNet(nn.Module, PreTrainedModelMixin):
    def __init__(self, config: AlexNetConfig) -> None:
        super().__init__()
        self.config = config

        self.conv = nn.Sequential(
            nn.Conv2d(config.in_channels, 64, kernel_size=11, stride=4, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )

        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(6, 6))
        hidden_1, hidden_2 = config.classifier_hidden_features
        self.fc = nn.Sequential(
            nn.Dropout(config.dropout),
            nn.Linear(256 * 6 * 6, hidden_1),
            nn.ReLU(),
            nn.Dropout(config.dropout),
            nn.Linear(hidden_1, hidden_2),
            nn.ReLU(),
            nn.Linear(hidden_2, config.num_classes),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)
        return x


@register_model
def alexnet(num_classes: int = 1000, **kwargs) -> AlexNet:
    config = AlexNetConfig(num_classes=num_classes, **kwargs)
    return AlexNet(config)



from pathlib import Path

from lucid.data.tokenizers import SpecialTokens
from lucid.models.text.bert import BERTTokenizerFast

__all__ = ["RoFormerTokenizerFast"]


class RoFormerTokenizerFast(BERTTokenizerFast):
    def __init__(
        self,
        vocab: dict[str, int] | None = None,
        vocab_file: Path | str | None = None,
        unk_token: SpecialTokens | str = SpecialTokens.UNK,
        pad_token: SpecialTokens | str = SpecialTokens.PAD,
        cls_token: str = "[CLS]",
        mask_token: str = "[MASK]",
        sep_token: str = "[SEP]",
        lowercase: bool = True,
        wordpieces_prefix: str = "##",
        max_input_chars_per_word: int = 100,
        clean_text: bool = True,
    ) -> None:
        super().__init__(
            vocab,
            vocab_file,
            unk_token,
            pad_token,
            cls_token,
            mask_token,
            sep_token,
            lowercase,
            wordpieces_prefix,
            max_input_chars_per_word,
            clean_text,
        )



import numpy as np
import torch

import lucid
from lucid.test._utils.test_utils import _UtilsTorchOpBase, _unwrap_callable
from lucid.test.core import TensorOpTorchCase


class _EinopsTorchBase(_UtilsTorchOpBase):
    _collect_nested_test_classes = True

    def _build_case(self) -> TensorOpTorchCase:
        forward_raw = getattr(type(self), "forward_op", None)
        torch_raw = getattr(type(self), "torch_forward", None)

        forward_op = _unwrap_callable(forward_raw)
        torch_forward = None if torch_raw is None else _unwrap_callable(torch_raw)

        return TensorOpTorchCase(
            name=self.case_name,
            build_inputs=lambda _: self._build_inputs(),
            forward_op=forward_op,
            expected_forward=self.expected_forward,
            expected_input_grads=self.expected_input_grads,
            requires_grad=self._normalize_bool_flags(
                self.requires_grad, len(self.inputs)
            ),
            check_finite=self.check_finite,
            rtol=self.rtol,
            atol=self.atol,
            torch_forward=torch_forward,
            compare_grad_with_torch=self.compare_grad_with_torch,
        )


class TestRearrange(_EinopsTorchBase):
    __test__ = False

    class TestTranspose(_EinopsTorchBase):
        case_name = "rearrange_transpose"
        input = np.arange(6, dtype=np.float64).reshape(2, 3)
        inputs = (input,)
        forward_op = staticmethod(lambda a: lucid.einops.rearrange(a, "b c -> c b"))
        torch_forward = staticmethod(lambda a: torch.permute(a, (1, 0)))
        expected_forward = input.T

    class TestMergeDims(_EinopsTorchBase):
        case_name = "rearrange_merge_dims"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.rearrange(a, "b c d -> b (c d)")
        )
        torch_forward = staticmethod(lambda a: torch.reshape(a, (2, 12)))
        expected_forward = input.reshape(2, 12)

    class TestSplitKnown(_EinopsTorchBase):
        case_name = "rearrange_split_known"
        input = np.arange(24, dtype=np.float64).reshape(6, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.rearrange(a, "(b c) d -> b c d", b=2, c=3)
        )
        torch_forward = staticmethod(lambda a: torch.reshape(a, (2, 3, 4)))
        expected_forward = input.reshape(2, 3, 4)

    class TestFlattenNamed(_EinopsTorchBase):
        case_name = "rearrange_flatten_named"
        input = np.arange(24, dtype=np.float64).reshape(6, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.rearrange(a, "(a b) c -> a b c", a=2, b=3)
        )
        torch_forward = staticmethod(lambda a: torch.reshape(a, (2, 3, 4)))
        expected_forward = input.reshape(2, 3, 4)

    class TestEllipsis(_EinopsTorchBase):
        case_name = "rearrange_ellipsis"
        input = np.arange(120, dtype=np.float64).reshape(2, 3, 4, 5)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.rearrange(a, "... c d -> d ... c")
        )
        torch_forward = staticmethod(lambda a: torch.permute(a, (3, 0, 1, 2)))
        expected_forward = np.transpose(input, (3, 0, 1, 2))

    class TestFlattenAll(_EinopsTorchBase):
        case_name = "rearrange_flatten_all"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.rearrange(a, "a b c -> (a b c)")
        )
        torch_forward = staticmethod(lambda a: torch.reshape(a, (24,)))
        expected_forward = input.reshape(-1)


class TestReduce(_EinopsTorchBase):
    __test__ = False

    class TestSumOverTwoAxes(_EinopsTorchBase):
        case_name = "reduce_sum_two_axes"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.reduce(a, "b c d -> b", reduction="sum")
        )
        torch_forward = staticmethod(lambda a: torch.sum(a, dim=(1, 2)))
        expected_forward = input.sum(axis=(1, 2))

    class TestMeanAxis(_EinopsTorchBase):
        case_name = "reduce_mean_axis"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.reduce(a, "b c d -> b d", reduction="mean")
        )
        torch_forward = staticmethod(lambda a: torch.mean(a, dim=1))
        expected_forward = input.mean(axis=1)

    class TestKeepAxis(_EinopsTorchBase):
        case_name = "reduce_keep_axis"
        input = np.arange(24, dtype=np.float64).reshape(6, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.reduce(a, "(b c) d -> b", reduction="mean", b=2, c=3)
        )
        torch_forward = staticmethod(
            lambda a: torch.mean(a.reshape(2, 3, 4), dim=(1, 2))
        )
        expected_forward = input.reshape(2, 3, 4).mean(axis=(1, 2))

    class TestTupleOutput(_EinopsTorchBase):
        case_name = "reduce_tuple_output"
        input = np.arange(48, dtype=np.float64).reshape(6, 8)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.reduce(
                a, "(b c) d -> (b d)", reduction="sum", b=2, c=3
            )
        )
        torch_forward = staticmethod(
            lambda a: torch.sum(a.reshape(2, 3, 8), dim=1).reshape(-1)
        )
        expected_forward = input.reshape(2, 3, 8).sum(axis=1).reshape(-1)

    class TestEllipsis(_EinopsTorchBase):
        case_name = "reduce_ellipsis"
        input = np.arange(30, dtype=np.float64).reshape(2, 3, 5)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.reduce(a, "... c -> ...", reduction="sum")
        )
        torch_forward = staticmethod(lambda a: torch.sum(a, dim=-1))
        expected_forward = input.sum(axis=2)

    class TestReduceAll(_EinopsTorchBase):
        case_name = "reduce_all"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.reduce(a, "a b c ->", reduction="sum")
        )
        torch_forward = staticmethod(lambda a: torch.sum(a))
        expected_forward = input.sum()


class TestRepeat(_EinopsTorchBase):
    __test__ = False

    class TestNewAxis(_EinopsTorchBase):
        case_name = "repeat_new_axis"
        input = np.arange(8, dtype=np.float64).reshape(2, 4)
        inputs = (input,)
        forward_op = staticmethod(lambda a: lucid.einops.repeat(a, "b c -> b c n", n=2))
        torch_forward = staticmethod(lambda a: a.unsqueeze(-1).repeat(1, 1, 2))
        expected_forward = np.repeat(input[:, :, None], repeats=2, axis=2)

    class TestTwoNewAxes(_EinopsTorchBase):
        case_name = "repeat_two_new_axes"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.repeat(a, "b c d -> b c d n m", n=1, m=2)
        )
        torch_forward = staticmethod(
            lambda a: a.unsqueeze(-1).unsqueeze(-1).repeat(1, 1, 1, 1, 2)
        )
        expected_forward = np.repeat(input[:, :, :, None, None], repeats=2, axis=4)

    class TestPermuteAndRepeat(_EinopsTorchBase):
        case_name = "repeat_permute_and_repeat"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.repeat(a, "b c d -> d b c n", n=2)
        )
        torch_forward = staticmethod(
            lambda a: torch.permute(a, (2, 0, 1)).unsqueeze(-1).repeat(1, 1, 1, 2)
        )
        expected_forward = np.repeat(
            np.transpose(input, (2, 0, 1))[:, :, :, None], repeats=2, axis=3
        )

    class TestEllipsis(_EinopsTorchBase):
        case_name = "repeat_ellipsis"
        input = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.repeat(a, "... c -> ... c n", n=2)
        )
        torch_forward = staticmethod(
            lambda a: a.unsqueeze(-1).repeat_interleave(2, dim=-1)
        )
        expected_forward = np.repeat(input[..., None], repeats=2, axis=3)

    class TestSingletonSplit(_EinopsTorchBase):
        case_name = "repeat_singleton_split"
        input = np.arange(8, dtype=np.float64).reshape(2, 4)
        inputs = (input,)
        forward_op = staticmethod(
            lambda a: lucid.einops.repeat(a, "b c -> b (c n)", n=2)
        )
        torch_forward = staticmethod(lambda a: a.unsqueeze(-1))
        expected_forward = input[:, :, None]


class TestEinsum(_EinopsTorchBase):
    __test__ = False

    class TestOuter(_EinopsTorchBase):
        case_name = "einsum_outer"
        left = np.arange(6, dtype=np.float64).reshape(2, 3)
        right = np.arange(8, dtype=np.float64).reshape(2, 4)
        inputs = (left, right)
        forward_op = staticmethod(lambda a, b: lucid.einops.einsum("bi,bj->bij", a, b))
        torch_forward = staticmethod(lambda a, b: torch.einsum("bi,bj->bij", a, b))
        expected_forward = np.einsum("bi,bj->bij", left, right)

    class TestMatMul(_EinopsTorchBase):
        case_name = "einsum_matmul"
        left = np.arange(12, dtype=np.float64).reshape(3, 4)
        right = np.arange(20, dtype=np.float64).reshape(4, 5)
        inputs = (left, right)
        forward_op = staticmethod(lambda a, b: lucid.einops.einsum("ij,jk->ik", a, b))
        torch_forward = staticmethod(lambda a, b: torch.einsum("ij,jk->ik", a, b))
        expected_forward = np.einsum("ij,jk->ik", left, right)

    class TestTrace(_EinopsTorchBase):
        case_name = "einsum_trace"
        input = np.arange(9, dtype=np.float64).reshape(3, 3)
        inputs = (input,)
        forward_op = staticmethod(lambda a: lucid.einops.einsum("ii->", a))
        torch_forward = staticmethod(lambda a: torch.einsum("ii->", a))
        expected_forward = np.einsum("ii->", input)
        compare_grad_with_torch = False

    class TestIdentity(_EinopsTorchBase):
        case_name = "einsum_identity"
        input = np.arange(6, dtype=np.float64).reshape(2, 3)
        inputs = (input,)
        forward_op = staticmethod(lambda a: lucid.einops.einsum("b c", a))
        torch_forward = staticmethod(lambda a: torch.einsum("b c", a))
        expected_forward = input
        compare_grad_with_torch = False

    class TestImplicitMatrix(_EinopsTorchBase):
        case_name = "einsum_implicit_matrix"
        left = np.arange(6, dtype=np.float64).reshape(2, 3)
        right = np.arange(8, dtype=np.float64).reshape(2, 4)
        inputs = (left, right)
        forward_op = staticmethod(lambda a, b: lucid.einops.einsum("bi,bj", a, b))
        torch_forward = staticmethod(lambda a, b: torch.einsum("bi,bj", a, b))
        expected_forward = np.einsum("bi,bj", left, right)

    class TestImplicitScalar(_EinopsTorchBase):
        case_name = "einsum_implicit_scalar"
        left = np.arange(6, dtype=np.float64).reshape(2, 3)
        right = np.arange(6, dtype=np.float64).reshape(2, 3)
        inputs = (left, right)
        forward_op = staticmethod(lambda a, b: lucid.einops.einsum("bi,bi", a, b))
        torch_forward = staticmethod(lambda a, b: torch.einsum("bi,bi", a, b))
        expected_forward = np.einsum("bi,bi", left, right)

    class TestChainImplicit(_EinopsTorchBase):
        case_name = "einsum_chain_implicit"
        left = np.arange(4, dtype=np.float64).reshape(2, 2)
        middle = np.arange(6, dtype=np.float64).reshape(2, 3)
        right = np.arange(6, dtype=np.float64).reshape(3, 2)
        inputs = (left, middle, right)
        forward_op = staticmethod(
            lambda a, b, c: lucid.einops.einsum("b i,i j,j k", a, b, c)
        )
        torch_forward = staticmethod(
            lambda a, b, c: torch.einsum("b i,i j,j k", a, b, c)
        )
        expected_forward = np.einsum("b i,i j,j k", left, middle, right)

    class TestBatchMatMul(_EinopsTorchBase):
        case_name = "einsum_batch_matmul"
        left = np.arange(24, dtype=np.float64).reshape(2, 3, 4)
        right = np.arange(16, dtype=np.float64).reshape(2, 4, 2)
        inputs = (left, right)
        forward_op = staticmethod(
            lambda a, b: lucid.einops.einsum("b i j,b j k->b i k", a, b)
        )
        torch_forward = staticmethod(
            lambda a, b: torch.einsum("b i j,b j k->b i k", a, b)
        )
        expected_forward = np.einsum("b i j,b j k->b i k", left, right)



import math
from typing import Any, Literal

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

__all__ = [
    "Unfold",
    "Conv1d",
    "Conv2d",
    "Conv3d",
    "ConvTranspose1d",
    "ConvTranspose2d",
    "ConvTranspose3d",
    "ConstrainedConv1d",
    "ConstrainedConv2d",
    "ConstrainedConv3d",
]


def _single_to_tuple(value: Any, times: int) -> tuple[Any, ...]:
    if isinstance(value, tuple):
        return value
    return (value,) * times


class Unfold(nn.Module):
    def __init__(
        self,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...],
        padding: int | tuple[int, ...],
        dilation: int | tuple[int, ...] = 1,
    ) -> None:
        super().__init__()
        self.kernel_size = _single_to_tuple(kernel_size, 2)
        self.stride = _single_to_tuple(stride, 2)
        self.padding = _single_to_tuple(padding, 2)
        self.dilation = _single_to_tuple(dilation, 2)

    def forward(self, input_: Tensor) -> Tensor:
        if input_.ndim != 4:
            raise ValueError(f"'nn.Unfold' only supports 4D-tensors. (i.e. images)")

        unfolded = F.unfold(
            input_, self.kernel_size, self.stride, self.padding, self.dilation
        )

        N, C, *spatial_dims = input_.shape
        out_dims = []
        for in_dim, p, d, k, s in zip(
            spatial_dims, self.padding, self.dilation, self.kernel_size, self.stride
        ):
            eff_k = d * (k - 1) + 1
            out_dim = math.floor((in_dim + 2 * p - eff_k) / s + 1)
            out_dims.append(out_dim)

        L = math.prod(out_dims)
        Ck = C * math.prod(self.kernel_size)
        return unfolded.reshape(N, Ck, L)


_PaddingStr = Literal["same", "valid"]


class _ConvNd(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...],
        padding: _PaddingStr | int | tuple[int, ...],
        dilation: int | tuple[int, ...],
        groups: int,
        bias: bool,
        *,
        D: int,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.groups = groups

        self.kernel_size = _single_to_tuple(kernel_size, D)
        self.stride = _single_to_tuple(stride, D)
        self.dilation = _single_to_tuple(dilation, D)

        if isinstance(padding, str):
            if padding == "same":
                self.padding = tuple(
                    (self.dilation[i] * (self.kernel_size[i] - 1)) // 2
                    for i in range(D)
                )
            elif padding == "valid":
                self.padding = (0,) * D
            else:
                raise ValueError(f"Unknown padding string: {padding}")
        else:
            self.padding = _single_to_tuple(padding, D)

        if groups <= 0:
            raise ValueError("groups must be a positive integer.")
        if in_channels % groups != 0:
            raise ValueError("in_channels must be divisible by groups.")
        if out_channels % groups != 0:
            raise ValueError("out_channels mube be divisible by groups.")

        weight_ = lucid.empty(out_channels, in_channels // groups, *self.kernel_size)
        self.weight = nn.Parameter(weight_)

        if bias:
            bias_ = lucid.empty((out_channels,))
            self.bias = nn.Parameter(bias_)
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.kaiming_uniform(self.weight)

        if self.bias is not None:
            fan_in, _ = nn.init._dist._calculate_fan_in_and_fan_out(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform(self.bias, -bound, bound)

    def extra_repr(self) -> str:
        s = (
            f"{self.in_channels}, {self.out_channels}, "
            f"kernel_size={self.kernel_size}, stride={self.stride}"
        )
        if self.padding != (0,) * len(self.padding):
            s += f", padding={self.padding}"
        if self.dilation != (1,) * len(self.dilation):
            s += f", dilation={self.dilation}"
        if self.groups != 1:
            s += f", groups={self.groups}"
        if self.bias is None:
            s += ", bias=False"

        return s


_ConstrainedMode = Literal[
    "none",
    "nonneg",
    "sum_to_one",
    "zero_mean",
    "nonneg_sum1",
    "unit_l2",
    "max_l2",
    "fixed_center",
]
_ConstraintEnforce = Literal["forward", "post_step"]


class _ConstrainedConvNd(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...],
        padding: _PaddingStr | int | tuple[int, ...],
        dilation: int | tuple[int, ...],
        groups: int,
        bias: bool,
        *,
        constraint: _ConstrainedMode = "none",
        enforce: _ConstraintEnforce = "forward",
        eps: float = 1e-12,
        max_l2: float | None = None,
        center_value: float = -1.0,
        neighbor_sum: float = 1.0,
        D: int,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=D,
        )

        self.constraint = constraint
        self.enforce = enforce
        self.eps = eps
        self.max_l2 = max_l2
        self.center_value = center_value
        self.neighbor_sum = neighbor_sum
        self._reduce_axis = tuple(range(-D, 0))

        valid_constraints = {
            "none",
            "nonneg",
            "sum_to_one",
            "zero_mean",
            "nonneg_sum1",
            "unit_l2",
            "max_l2",
            "fixed_center",
        }
        if constraint not in valid_constraints:
            raise ValueError(f"Unknown constraint: {constraint}")
        if enforce not in {"forward", "post_step"}:
            raise ValueError(f"Unknown enforce mode: {enforce}")
        if constraint == "max_l2" and (max_l2 is None or max_l2 <= 0):
            raise ValueError("`max_l2` must be positive when constraint='max_l2'.")

        if constraint == "fixed_center":
            if any(k % 2 == 0 for k in self.kernel_size):
                raise ValueError(
                    "fixed_center requires odd kernel sizes for all spatial dims."
                )

            center_mask = lucid.zeros(1, 1, *self.kernel_size, device=self.device)
            center_idx = (0, 0, *[k // 2 for k in self.kernel_size])
            center_mask[center_idx] = 1.0

            self.register_buffer("_center_mask", center_mask)
            self.register_buffer("_neighbor_mask", 1.0 - center_mask)
        else:
            self.register_buffer("_center_mask", None)
            self.register_buffer("_neighbor_mask", None)

        self._center_mask: nn.Buffer | None
        self._neighbor_mask: nn.Buffer | None

    def _sum_spatial(self, w: Tensor) -> Tensor:
        return lucid.sum(w, axis=self._reduce_axis, keepdims=True)

    def _normalize_sum(self, w: Tensor, target_sum: float) -> Tensor:
        return w / (self._sum_spatial(w) + self.eps) * target_sum

    def _l2_spatial(self, w: Tensor) -> Tensor:
        return lucid.sqrt(self._sum_spatial(w * w) + self.eps)

    def _apply_constraint(self, w: Tensor) -> Tensor:
        if self.constraint == "none":
            return w
        if self.constraint == "nonneg":
            return F.relu(w)
        if self.constraint == "sum_to_one":
            return self._normalize_sum(w, 1.0)
        if self.constraint == "zero_mean":
            return w - lucid.mean(w, axis=self._reduce_axis, keepdims=True)
        if self.constraint == "nonneg_sum1":
            return self._normalize_sum(F.relu(w), 1.0)
        if self.constraint == "unit_l2":
            return w / self._l2_spatial(w)
        if self.constraint == "max_l2":
            ratio = self.max_l2 / self._l2_spatial(w)
            return w * lucid.clip(ratio, min_value=None, max_value=1.0)
        if self.constraint == "fixed_center":
            center = self._center_mask * self.center_value
            neighbors = w * self._neighbor_mask
            neighbors = self._normalize_sum(neighbors, self.neighbor_sum)
            return neighbors + center

        raise RuntimeError(f"Unhandled constraint: {self.constraint}")

    def _constrained_weight(self) -> Tensor:
        if self.enforce == "forward":
            return self._apply_constraint(self.weight)
        return self.weight

    def project_(self) -> "_ConstrainedConvNd":
        projected = self._apply_constraint(self.weight)
        self.weight.data = projected.data
        return self

    def extra_repr(self) -> str:
        s = super().extra_repr()
        s += f", constraint={self.constraint}, enforce={self.enforce}, eps={self.eps}"
        if self.constraint == "max_l2":
            s += f", max_l2={self.max_l2}"
        if self.constraint == "fixed_center":
            s += (
                f", center_value={self.center_value}, "
                f"neighbor_sum={self.neighbor_sum}"
            )
        return s


class Conv1d(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=1,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=3)
        return F.conv1d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self.weight, self.bias)


class Conv2d(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=2,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=4)
        return F.conv2d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self.weight, self.bias)


class Conv3d(_ConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            D=3,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=5)
        return F.conv3d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self.weight, self.bias)


class ConstrainedConv1d(_ConstrainedConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
        *,
        constraint: _ConstrainedMode = "none",
        enforce: _ConstraintEnforce = "forward",
        eps: float = 1e-12,
        max_l2: float | None = None,
        center_value: float = -1.0,
        neighbor_sum: float = 1.0,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            constraint=constraint,
            enforce=enforce,
            eps=eps,
            max_l2=max_l2,
            center_value=center_value,
            neighbor_sum=neighbor_sum,
            D=1,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=3)
        return F.conv1d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self._constrained_weight(), self.bias)


class ConstrainedConv2d(_ConstrainedConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
        *,
        constraint: _ConstrainedMode = "none",
        enforce: _ConstraintEnforce = "forward",
        eps: float = 1e-12,
        max_l2: float | None = None,
        center_value: float = -1.0,
        neighbor_sum: float = 1.0,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            constraint=constraint,
            enforce=enforce,
            eps=eps,
            max_l2=max_l2,
            center_value=center_value,
            neighbor_sum=neighbor_sum,
            D=2,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=4)
        return F.conv2d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self._constrained_weight(), self.bias)


class ConstrainedConv3d(_ConstrainedConvNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
        *,
        constraint: _ConstrainedMode = "none",
        enforce: _ConstraintEnforce = "forward",
        eps: float = 1e-12,
        max_l2: float | None = None,
        center_value: float = -1.0,
        neighbor_sum: float = 1.0,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            constraint=constraint,
            enforce=enforce,
            eps=eps,
            max_l2=max_l2,
            center_value=center_value,
            neighbor_sum=neighbor_sum,
            D=3,
        )

    def _conv_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=5)
        return F.conv3d(
            input_, weight, bias, self.stride, self.padding, self.dilation, self.groups
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_forward(input_, self._constrained_weight(), self.bias)


class _ConvTransposeNd(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...],
        padding: _PaddingStr | int | tuple[int, ...],
        output_padding: int | tuple[int, ...],
        dilation: int | tuple[int, ...],
        groups: int,
        bias: bool,
        *,
        D: int,
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.groups = groups

        self.kernel_size = _single_to_tuple(kernel_size, D)
        self.stride = _single_to_tuple(stride, D)
        self.dilation = _single_to_tuple(dilation, D)
        self.output_padding = _single_to_tuple(output_padding, D)

        if isinstance(padding, str):
            if padding == "same":
                self.padding = tuple(
                    (self.dilation[i] * (self.kernel_size[i] - 1)) // 2
                    for i in range(D)
                )
            elif padding == "valid":
                self.padding = (0,) * D
            else:
                raise ValueError(f"Unknown padding string: {padding}")
        else:
            self.padding = _single_to_tuple(padding, D)

        if groups <= 0:
            raise ValueError("groups must be a positive integer.")
        if in_channels % groups != 0:
            raise ValueError("in_channels must be divisible by groups.")
        if out_channels % groups != 0:
            raise ValueError("out_channels must be divisible by groups.")

        weight_ = lucid.empty(in_channels, out_channels // groups, *self.kernel_size)
        self.weight = nn.Parameter(weight_)

        if bias:
            bias_ = lucid.empty((out_channels,))
            self.bias = nn.Parameter(bias_)
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.kaiming_uniform(self.weight)
        if self.bias is not None:
            fan_in, _ = nn.init._dist._calculate_fan_in_and_fan_out(self.weight)
            if fan_in != 0:
                bound = 1 / math.sqrt(fan_in)
                nn.init.uniform(self.bias, -bound, bound)

    def extra_repr(self) -> str:
        s = (
            f"{self.in_channels}, {self.out_channels}, "
            f"kernel_size={self.kernel_size}, stride={self.stride}"
        )
        if self.padding != (0,) * len(self.padding):
            s += f", padding={self.padding}"
        if self.output_padding != (0,) * len(self.output_padding):
            s += f", output_padding={self.output_padding}"
        if self.dilation != (1,) * len(self.dilation):
            s += f", dilation={self.dilation}"
        if self.groups != 1:
            s += f", groups={self.groups}"
        if self.bias is None:
            s += ", bias=False"

        return s


class ConvTranspose1d(_ConvTransposeNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        output_padding: int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            dilation,
            groups,
            bias,
            D=1,
        )

    def _conv_transpose_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=3)
        return F.conv_transpose1d(
            input_,
            weight,
            bias,
            self.stride,
            self.padding,
            self.output_padding,
            self.dilation,
            self.groups,
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_transpose_forward(input_, self.weight, self.bias)


class ConvTranspose2d(_ConvTransposeNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        output_padding: int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            dilation,
            groups,
            bias,
            D=2,
        )

    def _conv_transpose_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=4)
        return F.conv_transpose2d(
            input_,
            weight,
            bias,
            self.stride,
            self.padding,
            self.output_padding,
            self.dilation,
            self.groups,
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_transpose_forward(input_, self.weight, self.bias)


class ConvTranspose3d(_ConvTransposeNd):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        output_padding: int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        bias: bool = True,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            output_padding,
            dilation,
            groups,
            bias,
            D=3,
        )

    def _conv_transpose_forward(
        self, input_: Tensor, weight: Tensor, bias: Tensor | None
    ) -> Tensor:
        lucid._check_input_dim(input_, dim=5)
        return F.conv_transpose3d(
            input_,
            weight,
            bias,
            self.stride,
            self.padding,
            self.output_padding,
            self.dilation,
            self.groups,
        )

    def forward(self, input_: Tensor) -> Tensor:
        return self._conv_transpose_forward(input_, self.weight, self.bias)
